from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import state_ops
from tensorflow.python.framework import ops
from tensorflow.python.training import optimizer
import hyperchamber as hc
import inspect
from tensorflow.python.ops.gradients_impl import _hessian_vector_product

class CompetitiveOptimizer(optimizer.Optimizer):
  """https://github.com/devzhk/Implicit-Competitive-Regularization/blob/master/optimizers.py ACGD"""
  def __init__(self, learning_rate=0.001, decay=0.9, gan=None, config=None, use_locking=False, name="CompetitiveOptimizer", optimizer=None):
    super().__init__(use_locking, name)
    self._decay = decay
    self.gan = gan
    self.config = config
    self.name = name
    self.learning_rate = learning_rate
    self.optimizer = self.gan.create_optimizer(optimizer)
 
  def _prepare(self):
    super()._prepare()
    self.optimizer._prepare()

  def _create_slots(self, var_list):
    super()._create_slots(var_list)
    self.optimizer._create_slots(var_list)

  def _apply_dense(self, grad, var):
    return self.optimizer._apply_dense(grad, var)

  def apply_gradients(self, grads_and_vars, global_step=None, name=None):
    d_vars = []
    d_grads = []
    g_vars = []
    g_grads = []
    for grad,var in grads_and_vars:
        if var in self.gan.d_vars():
            d_vars += [var]
            d_grads += [grad]
        elif var in self.gan.g_vars():
            g_vars += [var]
            g_grads += [grad]
        else:
            raise("Couldn't find var in g_vars or d_vars")
    min_params = d_vars
    max_params = g_vars
    grad_x = g_grads
    grad_y = d_grads
    lr = self.learning_rate

    # best
    grad_x_rev = tf.gradients(self.gan.loss.sample[0], max_params)#, grad_ys=self.gan.loss.sample[0], stop_gradients=max_params)
    grad_y_rev = tf.gradients(self.gan.loss.sample[1], min_params)#, grad_ys=self.gan.loss.sample[1], stop_gradients=min_params)

    f = self.gan.loss.sample[0]
    g = self.gan.loss.sample[1]

    x_loss = self.gan.loss.sample[1]
    y_loss = self.gan.loss.sample[0]

    x_grads = grad_x
    y_grads = grad_y

    if self.config.sga_lambda == 0:
        rhs_x = grad_x
        rhs_y = grad_y
    else:
        if self.config.sga_t2:
            hyp_y = self.hessian_vector_product(f, max_params, min_params, [lr * _g for _g in grad_x_rev])
            hyp_x = self.hessian_vector_product(g, min_params, max_params, [lr * _g for _g in grad_y_rev])
        else:
            hyp_x = self.hvpvec([_g*lr for _g in grad_x], max_params, grad_x_rev)
            hyp_y = self.hvpvec([_g*lr for _g in grad_y], min_params, grad_y_rev)
        self.gan.add_metric('hyp_x', sum([ tf.reduce_mean(_p) for _p in hyp_x]))
        self.gan.add_metric('hyp_y', sum([ tf.reduce_mean(_p) for _p in hyp_y]))
        if self.config.neg:
            rhs_x = [g - (self.config.sga_lambda or lr)*hyp for g, hyp in zip(grad_x, hyp_x)]
        else:
            rhs_x = [g + (self.config.sga_lambda or lr)*hyp for g, hyp in zip(grad_x, hyp_x)]
        rhs_y = [g - (self.config.sga_lambda or lr)*hyp for g, hyp in zip(grad_y, hyp_y)]

    if self.config.con5:
        cg_x = self.mgeneral_conjugate_gradient(grad_x=rhs_y, grad_y=rhs_x, x_loss=x_loss, y_loss=y_loss,
                x_params=min_params, y_params=max_params, x=rhs_x, b=rhs_x, nsteps=(self.config.nsteps or 3),
                lr_x=self.learning_rate, lr_y=self.learning_rate)

        cg_y = self.mgeneral_conjugate_gradient(grad_x=rhs_x, grad_y=rhs_y, x_loss=y_loss, y_loss=x_loss,
                x_params=max_params, y_params=min_params, x=rhs_y, b=rhs_y, nsteps=(self.config.nsteps or 3),
                lr_x=self.learning_rate, lr_y=self.learning_rate)
    elif self.config.con6:
        cg_x = self.mgeneral_conjugate_gradient(grad_x=rhs_y, grad_y=rhs_x, x_loss=x_loss, y_loss=y_loss,
                x_params=min_params, y_params=max_params, b=rhs_x, nsteps=(self.config.nsteps or 3),
                lr_x=self.learning_rate, lr_y=self.learning_rate)

        cg_y = self.mgeneral_conjugate_gradient(grad_x=rhs_x, grad_y=rhs_y, x_loss=y_loss, y_loss=x_loss,
                x_params=max_params, y_params=min_params, b=rhs_y, nsteps=(self.config.nsteps or 3),
                lr_x=self.learning_rate, lr_y=self.learning_rate)

    elif self.config.con4:
        cg_x, cg_y = self.sum_ak(x_grads=x_grads, y_grads=y_grads, x_loss=y_loss, y_loss=x_loss, x_params=min_params, y_params=max_params, nsteps=(self.config.nsteps or 3), lr=self.learning_rate)
    else:
        cg_x, cg_y = self.sum_ak(x_grads=x_grads, y_grads=y_grads, x_loss=x_loss, y_loss=y_loss, x_params=min_params, y_params=max_params, nsteps=(self.config.nsteps or 3), lr=self.learning_rate)
    #cg_x = self.sum_ak(x_grads=y_grads, y_grads=x_grads, x_loss=y_loss, y_loss=x_loss, x_params=max_params, y_params=min_params, nsteps=(self.config.nsteps or 3), lr=self.learning_rate)
    self.gan.add_metric('cg_x', sum([ tf.reduce_mean(_p) for _p in cg_x]))
    self.gan.add_metric('cg_y', sum([ tf.reduce_mean(_p) for _p in cg_y]))

    #new_grad_x = [_c * _g for _c, _g in zip(cg_x, rhs_x)]
    #new_grad_y = [_c * _g for _c, _g in zip(cg_y, rhs_y)]
    new_grad_x = cg_x
    new_grad_y = cg_y

    new_grads = new_grad_y + new_grad_x
    print("grads", len(new_grads))

    all_vars = d_vars + g_vars
    print("allv", len(all_vars))
    new_grads_and_vars = list(zip(new_grads, all_vars)).copy()
    return self.optimizer.apply_gradients(new_grads_and_vars)

  def sum_ak(self, x, y, f, g, nsteps=1):
      for i in range(nsteps):
          grads = tf.gradients(g, x)
          grads2 = tf.gradients(grads, y)
          g = self.d2xy_dy(f=f, dy=grads2, y=y, x=x)
      return g
  
  
  def d2xy_dy(self, f, dy, y, x):
      grads = tf.gradients(f, y)
  
      elemwise_products = [
              math_ops.multiply(grad_elem, tf.stop_gradient(v_elem))
              for grad_elem, v_elem in zip(grads, dy)
              if grad_elem is not None
              ]
  
      result = tf.gradients(elemwise_products, x)
      return result
  
  def conjugate_gradient(self, grad_x, grad_y, x_params, y_params, lr_x, lr_y, x, nsteps=10):
      gy = grad_y
      for i in range(nsteps):
          gx = [_g*lr_x for _g in x]
          h_1_v = self.fwd_gradients(gx, y_params, stop_gradients=y_params)
          gy = [ _h_1_v + _g  for _h_1_v, _g in zip(h_1_v, gy)]
          h_2_v = self.fwd_gradients(gy, x_params, stop_gradients=x_params)
          x = [_h_2_v + _x for _h_2_v, _x in zip(h_2_v, x)]
      return x
  
  def conjugate_gradient2(self, grad_x, grad_y, x_params, y_params, lr_x, lr_y, x, nsteps=10):
      for i in range(nsteps):
          h_1_v = self.hvpvec([_g*lr_x for _g in grad_x], y_params, [lr_x * _p for _p in x])
          h_1 = [lr_y * v for v in h_1_v]
          h_2 = self.hvpvec([_g*lr_y for _g in grad_y], x_params, h_1)
          x = [_h_2 + _x for _h_2, _x in zip(h_2, x)]
      return x
  
  def hessian_vector_product(self, ys, xs, xs2, vs, grads=None):
    # Validate the input
      length = len(xs)
      if len(vs) != length:
          raise ValueError("xs and v must have the same length.")

      # First backprop
      if grads is None:
          grads = tf.gradients(ys, xs)

      assert len(grads) == length
      elemwise_products = [
              math_ops.multiply(grad_elem, tf.stop_gradient(v_elem))
              for grad_elem, v_elem in zip(grads, vs)
              if grad_elem is not None
              ]

      # Second backprop
      return tf.gradients(elemwise_products, xs2)



  def sum_ak(self, x_loss, y_loss, x_grads, y_grads, x_params, y_params, lr, nsteps=10):
    def adjust(v):
        return [_v * lr for _v in v]

    for i in range(nsteps):
        move1 = self.hessian_vector_product(x_loss, y_params, x_params, adjust(x_grads))
        y_grads = [_yg + _m1 for _yg, _m1 in zip(y_grads, move1)]
        #self.gan.add_metric('m1', sum([ tf.reduce_sum(tf.abs(_m1)) for _m1 in move1]))

        move2 = self.hessian_vector_product(y_loss, x_params, y_params, adjust(y_grads))
        x_grads = [_xg + _m2 for _xg, _m2 in zip(x_grads, move2)]
        #self.gan.add_metric('m2', sum([ tf.reduce_sum(tf.abs(_m2)) for _m2 in move2]))
    return x_grads, y_grads

  def mgeneral_conjugate_gradient(self, x_loss, y_loss, grad_x, grad_y, x_params, y_params, b, lr_x, lr_y, x=None, nsteps=10):
    if x is None:
        x = [tf.zeros_like(_b) for _b in b]
    eps = 1e-8
    r = [tf.identity(_b) for _b in b]
    p = [tf.identity(_r) for _r in r]
    rdotr = [self.dot(_r, _r) for _r in r]
    for i in range(nsteps):
        h_1_v = self.hessian_vector_product(x_loss, y_params, x_params, [lr_x * _p for _p in p])
        h_2_v = self.hessian_vector_product(y_loss, x_params, y_params, [lr_x * _h for _h in h_1_v])

        Avp_ = [_p + lr_x*_h_2 for _p, _h_2 in zip(p, h_2_v)]

        alpha = [_rdotr / (self.dot(_p, _avp_)+eps) for _rdotr, _p, _avp_ in zip(rdotr, p, Avp_)]
        x = [_alpha * _p for _alpha, _p in zip(alpha, p)]

        r = [_alpha * _avp_ for _alpha, _avp_ in zip(alpha, Avp_)]
        new_rdotr = [self.dot(_r, _r) for _r in r]
        beta = [_new_rdotr / (_rdotr+eps) for _new_rdotr, _rdotr in zip(new_rdotr, rdotr)]
        p = [_r + _beta * _p for _r, _beta, _p in zip(r,beta,p)]
        self.gan.add_metric('rdotr', sum([ tf.reduce_sum(tf.abs(_p)) for _p in rdotr]))

        rdotr = new_rdotr
    return x

  def dot(self, r, p):
    return r * p
    s = r.shape
    r = tf.reshape(r, [-1])
    p = tf.reshape(p, [-1])
    result = tf.tensordot(r, p, axes=0)
    return tf.reshape(result, s)

  def hvpvec(self, ys, xs, vs):
    #print("VS", len(vs), "XS", len(xs), "YS", len(ys))
    #return _hessian_vector_product(xs, ys, vs)
    #return tf.gradients(ys, xs, grad_ys=vs)
    #result = tf.gradients(ys, xs)
    #for r,v in zip(result, vs):
    #    print("R", r)
    #    print("V", v)
    result = self.fwd_gradients(ys, xs, stop_gradients=xs)
    #print("R", len(result))
    #print("V", len(vs))

    res = [ ]
    print("LEN", len(xs), len(result), len(ys))
    for y, r, v in zip(ys, result, vs):
        if r is not None:
            res += [v+r]
        else:
            print("[Warning] Null value in hvpvec ", v)
            res += [tf.zeros_like(y)]
    return res

  def fwd_gradients(self, ys, xs, grad_xs=None, stop_gradients=None, colocate_gradients_with_ops=True, us=None):
    if us is None:
        us = [tf.zeros_like(y) + float('nan') for y in ys]
    print("YS", len(ys), "US", len(us))
    dydxs = tf.gradients(ys, xs, grad_ys=us,stop_gradients=stop_gradients,colocate_gradients_with_ops=colocate_gradients_with_ops, unconnected_gradients='zero')
    dysdx = tf.gradients(dydxs, us, grad_ys=grad_xs, colocate_gradients_with_ops=colocate_gradients_with_ops, unconnected_gradients='zero')
    return dysdx

  def _apply_sparse(self, grad, var):
    raise NotImplementedError("Sparse gradient updates are not supported.")

  def variables(self):
      return super().variables() + self.optimizer.variables()



import importlib
import json
import numpy as np
import os
import sys
import time
import uuid
import copy

from hypergan.discriminators import *
from hypergan.distributions import *
from hypergan.generators import *
from hypergan.inputs import *
from hypergan.samplers import *
from hypergan.trainers import *

import hyperchamber as hc
from hyperchamber import Config
from hypergan.ops import TensorflowOps
import hypergan as hg

from hypergan.gan_component import ValidationException, GANComponent
from .standard_gan import StandardGAN

class DistributionFilteringGAN(StandardGAN):
    """
    On Stabilizing Generative Adversarial Training with Noise
    https://arxiv.org/pdf/1906.04612v1.pdf
    """
    def create(self):
        config = self.config

        self.latent = self.create_component(config.z_distribution or config.latent)
        self.uniform_distribution = self.latent

        z_shape = self.ops.shape(self.latent.sample)
        self.android_input = tf.reshape(self.latent.sample, [-1])

        direction, slider = self.create_controls(self.ops.shape(self.android_input))
        self.slider = slider
        self.direction = direction
        z = self.android_input + slider * direction
        z = tf.maximum(-1., z)
        z = tf.minimum(1., z)
        z = tf.reshape(z, z_shape)
        self.control_z = z

        self.generator = self.create_component(config.generator, name="generator", input=z)
        self.noise_generator = self.create_component((config.noise_generator or config.generator), name="noise_generator", input=z)

        #x, g = tf.concat([self.inputs.x, self.inputs.x + self.noise_generator.sample], axis=3), tf.concat([self.generator.sample, self.generator.sample + self.noise_generator.sample], axis=3)

        x1, g1 = self.inputs.x, self.generator.sample
        self.discriminator = self.create_component(config.discriminator, name="discriminator", input=tf.concat([x1,g1],axis=0))
        x2, g2 = self.inputs.x+self.noise_generator.sample, self.generator.sample+self.noise_generator.sample
        self.loss = self.create_component(config.loss, discriminator=self.discriminator)
        self.noise_discriminator = self.create_component(config.discriminator, name="discriminator", input=tf.concat([x2,g2],axis=0), reuse=True)
        noise_loss = self.create_component(config.loss, discriminator=self.noise_discriminator)
        self.loss.sample[0] += noise_loss.sample[0]
        self.loss.sample[1] += noise_loss.sample[1]
        self.losses = [self.loss]
        self.trainer = self.create_component(config.trainer)

        self.android_output = tf.reshape(self.generator.sample, [-1])

    def g_vars(self):
        return self.latent.variables() + self.generator.variables() + self.noise_generator.variables()
    def d_vars(self):
        return self.discriminator.variables()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import state_ops
from tensorflow.python.framework import ops
from tensorflow.python.training import optimizer
import hyperchamber as hc
import inspect
from tensorflow.python.ops.gradients_impl import _hessian_vector_product

class CompetitiveOptimizer(optimizer.Optimizer):
  """https://github.com/devzhk/Implicit-Competitive-Regularization/blob/master/optimizers.py ACGD"""
  def __init__(self, learning_rate=0.001, decay=0.9, gan=None, config=None, use_locking=False, name="CompetitiveOptimizer", optimizer=None):
    super().__init__(use_locking, name)
    self._decay = decay
    self.gan = gan
    self.config = config
    self.name = name
    self.learning_rate = learning_rate
    self.optimizer = self.gan.create_optimizer(optimizer)
 
  def _prepare(self):
    super()._prepare()
    self.optimizer._prepare()

  def _create_slots(self, var_list):
    super()._create_slots(var_list)
    self.optimizer._create_slots(var_list)

  def _apply_dense(self, grad, var):
    return self.optimizer._apply_dense(grad, var)

  def apply_gradients(self, grads_and_vars, global_step=None, name=None):
    d_vars = []
    d_grads = []
    g_vars = []
    g_grads = []
    for grad,var in grads_and_vars:
        if var in self.gan.d_vars():
            d_vars += [var]
            d_grads += [grad]
        elif var in self.gan.g_vars():
            g_vars += [var]
            g_grads += [grad]
        else:
            raise("Couldn't find var in g_vars or d_vars")
    min_params = d_vars
    max_params = g_vars
    grad_x = g_grads
    grad_y = d_grads
    lr = self.learning_rate

    # best
    grad_x_rev = tf.gradients(self.gan.loss.sample[0], max_params)#, grad_ys=self.gan.loss.sample[0], stop_gradients=max_params)
    grad_y_rev = tf.gradients(self.gan.loss.sample[1], min_params)#, grad_ys=self.gan.loss.sample[1], stop_gradients=min_params)

    f = self.gan.loss.sample[0]
    g = self.gan.loss.sample[1]
    
    #hyp_x = self.hvpvec(grad_y, max_params, grad_x_rev)
    #hyp_y = self.hvpvec(grad_x, min_params, grad_y_rev)

    # best
    #hyp_x = self.hvpvec(self.gan.loss.sample[0], max_params, grad_x_rev)
    #hyp_y = self.hvpvec(self.gan.loss.sample[1], min_params, grad_y_rev)

    # pytorch
    #hyp_x = self.hvpvec(grad_y, max_params, [_g*lr for _g in grad_y])
    #hyp_y = self.hvpvec(grad_x, min_params, [_g*lr for _g in grad_x])

    #hyp_x = self.hvpvec(self.gan.loss.sample[0], max_params, grad_x_rev)
    #hyp_y = self.hvpvec(self.gan.loss.sample[1], min_params, grad_y_rev)

    #hyp_x = self.hvpvec(grad_x, max_params, grad_x_rev)
    #hyp_y = self.hvpvec(grad_y, min_params, grad_y_rev)
    hyp_x = self.hvpvec([_g*lr for _g in grad_x], max_params, grad_x_rev)
    hyp_y = self.hvpvec([_g*lr for _g in grad_y], min_params, grad_y_rev)
    # explicit
    #dy_g = tf.gradients(g, min_params)#, grad_ys=self.gan.loss.sample[0], stop_gradients=max_params)
    #dx_f = tf.gradients(f, max_params)#, grad_ys=self.gan.loss.sample[1], stop_gradients=min_params)
    #dy_g = [_g * lr for _g in dy_g]
    #dx_f = [_g * lr for _g in dx_f]
    #hyp_x = self.d2xy_dy(f=f, dy=dy_g, y=min_params, x=max_params)
    #hyp_y = self.d2xy_dy(f=g, dy=dx_f, y=max_params, x=min_params)

    #hyp_x = self.hvpvec(grad_x, max_params, grad_x_rev)
    #hyp_y = self.hvpvec(grad_y, min_params, grad_y_rev)

    #hyp_x = self.hvpvec(self.gan.loss.sample[0], max_params, scaled_grad_x)
    #hyp_y = self.hvpvec(self.gan.loss.sample[1], min_params, scaled_grad_y)

    #hyp_x = tf.gradients(grad_y, max_params)
    #hyp_x = [_h * _g for _h, _g in zip(hyp_x, scaled_grad_x)]
    #hyp_y = tf.gradients(grad_x, min_params)
    #hyp_y = [_h * _g for _h, _g in zip(hyp_y, scaled_grad_y)]
    self.gan.add_metric('hyp_x', sum([ tf.reduce_mean(_p) for _p in hyp_x]))
    self.gan.add_metric('hyp_y', sum([ tf.reduce_mean(_p) for _p in hyp_y]))
    if self.config.sga_lambda == 0:
        rhs_x = grad_x
        rhs_y = grad_y
    else:
        if self.config.neg:
            rhs_x = [g - (self.config.sga_lambda or lr)*hyp for g, hyp in zip(grad_x, hyp_x)]
        else:
            rhs_x = [g + (self.config.sga_lambda or lr)*hyp for g, hyp in zip(grad_x, hyp_x)]
        rhs_y = [g - lr*hyp for g, hyp in zip(grad_y, hyp_y)]
    #old_y = [tf.zeros_like(_d * tf.math.sqrt(self.learning_rate)) for _d in d_grads]
    #
    #p_y2 = [_c * tf.math.sqrt(self.learning_rate) for _c in p_y]


    #------------
    if self.config.conjugate:
        cg_y = self.mgeneral_conjugate_gradient(grad_x=grad_y, grad_y=grad_y_rev,
                x_params=min_params, y_params=max_params, b=rhs_y, x=None, nsteps=(self.config.nsteps or 3),
                lr_x=self.learning_rate, lr_y=self.learning_rate)

        #cg_y = [c * lr for c in cg_y]
        #cg_x = [c * lr for c in cg_x]
        #rhs_x = cg_x
        #hcg = self.hvpvec(grad_y, max_params, cg_y)
        #hcg = [lr * _p + _g for _p,_g in zip(hcg, rhs_x)]

        rhs_y = cg_y

        #rhs_x = [_cg_x * _rhs_x for _cg_x, _rhs_x in zip(cg_x, rhs_x)]
        #rhs_y = [_cg_y * _rhs_y for _cg_y, _rhs_y in zip(cg_y, rhs_y)]
    #-----
    if self.config.conjugate3:
        cg_y = self.conjugate_gradient(grad_x=grad_y, grad_y=grad_y_rev,
                x_params=min_params, y_params=max_params, x=rhs_y, nsteps=(self.config.nsteps or 3),
                lr_x=self.learning_rate, lr_y=self.learning_rate)

        cg_x = self.conjugate_gradient(grad_x=grad_x, grad_y=grad_x_rev,
                x_params=max_params, y_params=min_params, x=rhs_x, nsteps=(self.config.nsteps or 3),
                lr_x=self.learning_rate, lr_y=self.learning_rate)
        if self.config.sum == "1-":
            rhs_x = [(1.0-_cg_x) * _rhs_x for _cg_x, _rhs_x in zip(cg_x, rhs_x)]
            rhs_y = [(1.0-_cg_y) * _rhs_y for _cg_y, _rhs_y in zip(cg_y, rhs_y)]
        if self.config.sum == "1+":
            rhs_x = [(1.0+_cg_x) * _rhs_x for _cg_x, _rhs_x in zip(cg_x, rhs_x)]
            rhs_y = [(1.0+_cg_y) * _rhs_y for _cg_y, _rhs_y in zip(cg_y, rhs_y)]
        elif self.config.sum == "=":
            #rhs_x = [(1.0-_cg_x) * _rhs_x for _cg_x, _rhs_x in zip(cg_x, rhs_x)]
            #rhs_y = [(1.0-_cg_y) * _rhs_y for _cg_y, _rhs_y in zip(cg_y, rhs_y)]
            rhs_x = cg_x
            rhs_y = cg_y
        else:
            rhs_x = [_cg_x * _rhs_x for _cg_x, _rhs_x in zip(cg_x, rhs_x)]
            rhs_y = [_cg_y * _rhs_y for _cg_y, _rhs_y in zip(cg_y, rhs_y)]
        self.gan.add_metric('cg_x', sum([ tf.reduce_mean(_p) for _p in cg_x]))
        self.gan.add_metric('cg_y', sum([ tf.reduce_mean(_p) for _p in cg_y]))

    if self.config.conjugate5:
        cg_y = self.conjugate_gradient(grad_x=rhs_y, grad_y=grad_y_rev,
                x_params=min_params, y_params=max_params, x=rhs_y, nsteps=(self.config.nsteps or 3),
                lr_x=self.learning_rate, lr_y=self.learning_rate)

        cg_x = self.conjugate_gradient(grad_x=rhs_x, grad_y=grad_x_rev,
                x_params=max_params, y_params=min_params, x=rhs_x, nsteps=(self.config.nsteps or 3),
                lr_x=self.learning_rate, lr_y=self.learning_rate)

        rhs_x = cg_x
        rhs_y = cg_y
        self.gan.add_metric('cg_x', sum([ tf.reduce_mean(_p) for _p in cg_x]))
        self.gan.add_metric('cg_y', sum([ tf.reduce_mean(_p) for _p in cg_y]))

    if self.config.conjugate6:
        cg_y = self.conjugate_gradient2(grad_x=rhs_y, grad_y=grad_y_rev,
                x_params=min_params, y_params=max_params, x=rhs_y, nsteps=(self.config.nsteps or 3),
                lr_x=self.learning_rate, lr_y=self.learning_rate)

        cg_x = self.conjugate_gradient2(grad_x=rhs_x, grad_y=grad_x_rev,
                x_params=max_params, y_params=min_params, x=rhs_x, nsteps=(self.config.nsteps or 3),
                lr_x=self.learning_rate, lr_y=self.learning_rate)

        rhs_x = cg_x
        rhs_y = cg_y
        self.gan.add_metric('cg_x', sum([ tf.reduce_mean(_p) for _p in cg_x]))
        self.gan.add_metric('cg_y', sum([ tf.reduce_mean(_p) for _p in cg_y]))

    if self.config.conjugate8:
        cg_y = self.conjugate_gradient2(grad_x=grad_y, grad_y=grad_y_rev,
                x_params=min_params, y_params=max_params, x=rhs_y, nsteps=(self.config.nsteps or 3),
                lr_x=self.learning_rate, lr_y=self.learning_rate)

        cg_x = self.conjugate_gradient2(grad_x=grad_x, grad_y=grad_x_rev,
                x_params=max_params, y_params=min_params, x=rhs_x, nsteps=(self.config.nsteps or 3),
                lr_x=self.learning_rate, lr_y=self.learning_rate)

        rhs_x = [(1.0+_cg_x) * _rhs_x for _cg_x, _rhs_x in zip(cg_x, rhs_x)]
        rhs_y = [(1.0+_cg_y) * _rhs_y for _cg_y, _rhs_y in zip(cg_y, rhs_y)]
        self.gan.add_metric('cg_x', sum([ tf.reduce_mean(_p) for _p in cg_x]))
        self.gan.add_metric('cg_y', sum([ tf.reduce_mean(_p) for _p in cg_y]))



    if self.config.conjugate7:
        cg_y = self.conjugate_gradient(grad_x=rhs_y, grad_y=grad_y_rev,
                x_params=min_params, y_params=max_params, x=rhs_y, nsteps=(self.config.nsteps or 3),
                lr_x=self.learning_rate, lr_y=self.learning_rate)

        cg_x = self.conjugate_gradient(grad_x=rhs_x, grad_y=grad_x_rev,
                x_params=max_params, y_params=min_params, x=rhs_x, nsteps=(self.config.nsteps or 3),
                lr_x=self.learning_rate, lr_y=self.learning_rate)

        rhs_x = cg_x
        rhs_y = cg_y
        self.gan.add_metric('cg_x', sum([ tf.reduce_mean(_p) for _p in cg_x]))
        self.gan.add_metric('cg_y', sum([ tf.reduce_mean(_p) for _p in cg_y]))






    if self.config.conjugate2:
        cg_x = self.mgeneral_conjugate_gradient(grad_x=grad_x, grad_y=grad_x_rev,
                x_params=max_params, y_params=min_params, b=rhs_x, x=None, nsteps=(self.config.nsteps or 3),
                lr_x=self.learning_rate, lr_y=self.learning_rate)
        rhs_x = cg_x
    if self.config.sum_ak:
        ak_x = self.sum_ak(x=x_params, y=y_params, f=f, g=g)
        ak_y = self.sum_ak(x=y_params, y=x_params, f=g, g=f)
        new_grad_y = [_ak * _rhs for _ak, _rhs in zip(ak_y, rhs_y)]
        new_grad_x = [_ak * _rhs for _ak, _rhs in zip(ak_x, rhs_x)]

        hcg_y = self.hvpvec(grad_x, min_params, cg_x)
        new_grad_y = [g+hcg for g, hcg in zip(grad_y, hcg_y)]

    

    new_grad_x = rhs_x
    new_grad_y = rhs_y

    new_grads = new_grad_y + new_grad_x

    all_vars = d_vars + g_vars
    new_grads_and_vars = list(zip(new_grads, all_vars)).copy()
    return self.optimizer.apply_gradients(new_grads_and_vars)

  def sum_ak(self, x, y, f, g, nsteps=1):
    for i in range(nsteps):
        grads = tf.gradients(g, x)
        grads2 = tf.gradients(grads, y)
        g = self.d2xy_dy(f=f, dy=grads2, y=y, x=x)
    return g


  def d2xy_dy(self, f, dy, y, x):
    grads = tf.gradients(f, y)
  
    elemwise_products = [
        math_ops.multiply(grad_elem, tf.stop_gradient(v_elem))
        for grad_elem, v_elem in zip(grads, dy)
        if grad_elem is not None
    ]
  
    result = tf.gradients(elemwise_products, x)
    return result

  def conjugate_gradient(self, grad_x, grad_y, x_params, y_params, lr_x, lr_y, x, nsteps=10):
    gy = grad_y
    for i in range(nsteps):
        gx = [_g*lr_x for _g in x]
        h_1_v = self.fwd_gradients(gx, y_params, stop_gradients=y_params)
        gy = [ _h_1_v + _g  for _h_1_v, _g in zip(h_1_v, gy)]
        h_2_v = self.fwd_gradients(gy, x_params, stop_gradients=x_params)
        x = [_h_2_v + _x for _h_2_v, _x in zip(h_2_v, x)]
    return x

  def conjugate_gradient2(self, grad_x, grad_y, x_params, y_params, lr_x, lr_y, x, nsteps=10):
    for i in range(nsteps):
        h_1_v = self.hvpvec([_g*lr_x for _g in grad_x], y_params, [lr_x * _p for _p in x])
        h_1 = [lr_y * v for v in h_1_v]
        h_2 = self.hvpvec([_g*lr_y for _g in grad_y], x_params, h_1)
        x = [_h_2 + _x for _h_2, _x in zip(h_2, x)]
    return x



  def mgeneral_conjugate_gradient(self, grad_x, grad_y, x_params, y_params, b, lr_x, lr_y, x=None, nsteps=10):
    if x is None:
        x = [tf.zeros_like(_b) for _b in b]
    eps = 1e-8
    r = [tf.identity(_b) for _b in b]
    p = [tf.identity(_r) for _r in r]
    rdotr = [self.dot(_r, _r) for _r in r]
    for i in range(nsteps):
        #self.gan.add_metric("hp", sum([ tf.reduce_mean(_p) for _p in lr_x_mul_p]))
        h_1_v = self.hvpvec([_g*lr_x for _g in grad_x], y_params, [lr_x * _p for _p in p])
        h_1 = [lr_y * v for v in h_1_v]
        #self.gan.add_metric("h_1", sum([ tf.reduce_mean(_p) for _p in h_1]))
        h_2_v = self.hvpvec([_g*lr_y for _g in grad_y], x_params, h_1)
        h_2 = [lr_x * v for v in h_2_v]
        #self.gan.add_metric("h_2", sum([ tf.reduce_mean(_p) for _p in h_2]))
        Avp_ = [_p + _h_2 for _p, _h_2 in zip(p, h_2)]
        alpha = [_rdotr / (self.dot(_p, _avp_)+eps) for _rdotr, _p, _avp_ in zip(rdotr, p, Avp_)]
        x = [_alpha * _p for _alpha, _p in zip(alpha, p)]

        r = [_alpha * _avp_ for _alpha, _avp_ in zip(alpha, Avp_)]
        new_rdotr = [self.dot(_r, _r) for _r in r]
        beta = [_new_rdotr / (_rdotr+eps) for _new_rdotr, _rdotr in zip(new_rdotr, rdotr)]
        p = [_r + _beta * _p for _r, _beta, _p in zip(r,beta,p)]
        rdotr = new_rdotr
    return x

  def dot(self, r, p):
    return r * p
    s = r.shape
    r = tf.reshape(r, [-1])
    p = tf.reshape(p, [-1])
    result = tf.tensordot(r, p, axes=0)
    return tf.reshape(result, s)

  def hvpvec(self, ys, xs, vs):
    #print("VS", len(vs), "XS", len(xs), "YS", len(ys))
    #return _hessian_vector_product(ys, xs, vs)
    #result = tf.gradients(ys, xs)
    #for r,v in zip(result, vs):
    #    print("R", r)
    #    print("V", v)
    result = self.fwd_gradients(ys, xs, stop_gradients=xs)
    #print("R", len(result))
    #print("V", len(vs))
    result = [ r * v for r, v in zip(result, vs) ]
    return result

  def fwd_gradients(self, ys, xs, grad_xs=None, stop_gradients=None, colocate_gradients_with_ops=True, us=None):
    if us is None:
        us = [tf.zeros_like(y) + float('nan') for y in ys]
    print("YS", len(ys), "US", len(us))
    dydxs = tf.gradients(ys, xs, grad_ys=us,stop_gradients=stop_gradients,colocate_gradients_with_ops=colocate_gradients_with_ops)
    dysdx = tf.gradients(dydxs, us, grad_ys=grad_xs, colocate_gradients_with_ops=colocate_gradients_with_ops)
    return dysdx

  def _apply_sparse(self, grad, var):
    raise NotImplementedError("Sparse gradient updates are not supported.")

  def variables(self):
      return super().variables() + self.optimizer.variables()



#From https://gist.github.com/EndingCredits/b5f35e84df10d46cfa716178d9c862a3
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from hypergan.gan_component import ValidationException
import hyperchamber as hc
import numpy as np
import inspect
from operator import itemgetter
from hypergan.train_hooks.base_train_hook import BaseTrainHook

class RollingMemoryTrainHook(BaseTrainHook):
  "Keeps a rolling memory of the best scoring discriminator samples."
  def __init__(self, gan=None, config=None, trainer=None, name="RollingMemoryTrainHook"):
    super().__init__(config=config, gan=gan, trainer=trainer, name=name)
    self.train_hook_index = len(trainer.train_hooks)
    config = hc.Config(config)
    s = self.gan.ops.shape(self.gan.generator.sample)
    self.memories = {}
    for pairs in self.config.types:
        for _type in pairs.split("/"):
            if _type in self.memories:
                pass
            elif _type == "g":
                self.memories["g"]={"d_input": self.gan.generator.sample}
            elif _type == "x":
                self.memories["x"]={"d_input": self.gan.inputs.x}
            elif _type[0:4] == "g(mz":
                src = self.source(_type)
                with tf.variable_scope((self.config.name or self.name), reuse=self.gan.reuse) as scope:
                    mem=tf.get_variable(self.gan.ops.generate_name()+"_z_dontsave", src.shape, dtype=tf.float32,
                              initializer=tf.compat.v1.constant_initializer(0), 
                              aggregation=tf.VariableAggregation.ONLY_FIRST_REPLICA, 
                              trainable=False, 
                              synchronization=tf.VariableSynchronization.NONE)
                mem_gen = self.gan.create_component(self.gan.config.generator, name='generator', input=mem, reuse=True).sample
                gen = self.gan.generator.sample
                mem_sw = tf.reshape(self.sw(mem_gen, _type), [self.gan.batch_size(), 1])
                self.memories[_type]={"var": mem, "source": src, "sw": mem_sw,
                                      "d_input": mem_gen,
                                      "assign": tf.assign(mem, self.select_top(src, gen, _type) * mem_sw + (1.0 - mem_sw) * mem)
                                      }
            else:
                src = self.source(_type)
                with tf.variable_scope((self.config.name or self.name), reuse=self.gan.reuse) as scope:
                    mem=tf.get_variable(self.gan.ops.generate_name()+"_dontsave", src.shape, dtype=tf.float32,
                              initializer=tf.compat.v1.constant_initializer(0), 
                              aggregation=tf.VariableAggregation.ONLY_FIRST_REPLICA, 
                              trainable=False, 
                              synchronization=tf.VariableSynchronization.NONE)
                self.memories[_type]={"var": mem, "source": src, "sw": self.sw(src, _type),
                                      "d_input": mem,
                                      "assign": tf.assign(mem, self.select_top(src, src, _type) * self.sw(mem, _type) + (1.0 - self.sw(mem, _type)) * mem)
                                      }
    self.loss = [tf.zeros(1), tf.zeros(1)]
    for i, pair in enumerate(self.config.types):
        left, right = pair.split("/")
        vleft = self.memories[left]["d_input"]
        vright = self.memories[right]["d_input"]
        d = gan.create_component(gan.config.discriminator, name="discriminator", input=tf.concat([vleft, vright],axis=0), features=[gan.features], reuse=True)
        l = gan.create_component(gan.config.loss, discriminator=d)
        if self.config.add_to_losses:
            self.gan.losses += [l]
        if self.config.lambdas and i < len(self.config.lambdas):
            lam = self.config.lambdas[i]
        else:
            lam = (self.config.lam or 1.0)
        self.loss[0] += lam  * l.sample[0]
        self.loss[1] += lam * l.sample[1]
        self.gan.add_metric("rolling_d_loss_"+pair, self.loss[0])
    self.update_memory = []
    for _type, memory in self.memories.items():
      if "assign" in memory:
        self.update_memory += [memory["assign"]]
    self.update_memory = tf.group(*self.update_memory)

  def source(self, name):
    name = name.replace("+","").replace("-","")
    if name == "mg":
      return self.gan.generator.sample
    if name == "mx":
      return self.gan.inputs.x
    if name[0:4] == "g(mz":
      return self.gan.latent.sample
    raise ValidationException("Unknown rolling type: " + name)

  def select_top(self, src, var, name):
    if (self.config.top_k or 1) != 1:
      return var#unsupported
    print("VAR ", var, "SRNC", src, self.sw(var, name))
    if self.config.reverse_top:
        if "+" in name:
            name = name.replace('+','-')
        else:
            name = name.replace('-','+')

    sw = self.sw(var, name)
    sw = tf.reshape(sw, [self.ops.shape(src)[0]] + [1 for i in range(len(self.ops.shape(src))-1)])
    top = src * sw
    top = tf.reduce_sum(top, axis=[0], keep_dims=True)
    top = tf.tile(top, [self.ops.shape(src)[0]] + [1 for i in range(len(self.ops.shape(src))-1)])
    return top

  def sw(self, var, name):
    def calculate(dscore):
      swx = dscore
      swx = tf.reshape(swx, [-1])
      _, swx = tf.nn.top_k(swx, k=(self.config.top_k or 1), sorted=True, name=None)
      swx = tf.one_hot(swx, self.gan.batch_size(), dtype=tf.float32)
      swx = tf.reduce_sum(swx, reduction_indices=0)
      swx = tf.reshape(swx, [self.gan.batch_size(), 1, 1, 1])
      return swx
    def create_disc(var):
      return self.gan.create_component(self.gan.config.discriminator, name="discriminator", input=tf.concat([var, var],axis=0), features=[self.gan.features], reuse=True)

    d = create_disc(var)
    l = self.gan.create_component(self.gan.config.loss, discriminator=d)

    if name == "mg-":
        return calculate(-l.d_fake)
    if name == "mg+":
        return calculate(l.d_fake)
    if name == "mx-":
        return calculate(-l.d_real)
    if name == "mx+":
        return calculate(l.d_real)

    if name == "g(mz-)":
        return calculate(-l.d_fake)
    if name == "g(mz+)":
        return calculate(l.d_fake)


    raise ValidationException("Unknown rolling type: " + name)

  def distributed_step(self, input_iterator_next):
    def assign_m(name):
        op=self.gan.replica.trainer.train_hooks[self.train_hook_index].memories[name]["assign"]
        with tf.control_dependencies([op]):
            return tf.no_op()
        #return mg.assign(gen * swg + (1.0 - swg) * mg)
    #mxop = self.gan.distribution_strategy.extended.update(self.mx, assign_mx, args=(self.swx,))
    #mgop = self.gan.distribution_strategy.extended.update(self.mg, assign_mg, args=(self.gan.generator.sample, self.swg,))
    ops = []
    for name, memory in self.memories.items():
        if name[0] == "m":
            ops.append(self.gan.distribution_strategy.extended.call_for_each_replica(assign_m, args=(name,)))
    return ops

  def distributed_debug(self):
    ops = []
    for name, memory in self.memories.items():
        ops.append(self.gan.distribution_strategy.extended.read_var(memory["var"]))
    return ops


  def distributed_initial_step(self, input_iterator_next):
    def assign_mx(mx, inp):
        return mx.assign(inp)

    def assign_mg(mg, gen):
        return mg.assign(gen)

    ops = []
    for name, memory in self.memories.items():
        op = None
        if "mx" in name:
            op = self.gan.distribution_strategy.extended.call_for_each_replica(assign_mx, args=(memory["var"],input_iterator_next,))
        elif "mg" in name:
            op = self.gan.distribution_strategy.extended.call_for_each_replica(assign_mg, args=(memory["var"],self.gan.generator.sample,))
        if op is not None:
            ops += [op]

    #mxop = self.gan.distribution_strategy.extended.update(self.mx, assign_mx)
    #mgop = self.gan.distribution_strategy.extended.update(self.mg, assign_mg, args=(self.gan.generator.sample,))
    return ops


  #def update_op(self):
  #    return tf.group(self.assign_mg, self.assign_mx)

  def before_step(self, step, feed_dict):
      if step == 0:
          for _type, mem in self.memories.items():
              if "var" in mem and "source" in mem:
                  self.gan.session.run(tf.assign(mem["var"], mem["source"]))
          #self.gan.session.run(tf.assign(self.mx, self.gan.inputs.x))
          #self.gan.session.run(tf.assign(self.mg, self.gan.generator.sample))

  def after_step(self, step, feed_dict):
      if (step % (self.config.steps_between_memory_update or 1)) == (self.config.offset_memory_update or 0):
          self.gan.session.run(self.update_memory)

  def variables(self):
    var = []
    for _type, memory in self.memories.items():
      if "var" in memory and "assign" in memory:
        var += [memory["var"]]
    return var

  def d_inputs(self):
    var = []
    for pairs in self.config.types:
      for _type in pairs.split("/"):
        if _type in self.memories:
          var += [self.memories[_type]["d_input"]]
    return var


  def losses(self):
    return self.loss



from hypergan.gans.standard_gan import StandardGAN
from hypergan.gan_component import GANComponent

def gan_factory(*args, **kw_args):
    if 'config' in kw_args:
        config = kw_args['config']
    elif len(args) > 0:
        config = args[0]
    else:
        config = None
    if config and 'class' in config:
        return GANComponent.lookup_function(None, config['class'])(*args, **kw_args)
    else:
        return StandardGAN(*args, **kw_args)

GAN=gan_factory



import math
import torch
import torch.nn as nn

import torch.nn.functional as F

class ScaledConv2d(nn.Module):
    def __init__(
            self,
            in_channel,
            out_channel,
            kernel_size,
            style_dim,
            padding=0,
            demodulate=True,
            upsample=True,
            downsample=False,
            lr_mul=1.0,
            blur_kernel=[1, 3, 3, 1],
            ):

        super(ScaledConv2d, self).__init__()
        self.out_channel = out_channel
        self.kernel_size = kernel_size
        fan_in = in_channel * kernel_size ** 2
        self.scale = 1 / math.sqrt(fan_in)
        self.padding = padding#kernel_size // 2
        self.weight = nn.Parameter(
            torch.randn(out_channel, in_channel, kernel_size, kernel_size)
        )

    def forward(self, input):
        batch, in_channel, height, width = input.shape
        weight = self.scale * self.weight
        out = F.conv2d(input, weight, padding=self.padding)

        return out




from hyperchamber import Config
from hypergan.gan_component import ValidationException, GANComponent
from hypergan.samplers.aligned_sampler import AlignedSampler
from hypergan.samplers.batch_sampler import BatchSampler
from hypergan.samplers.batch_walk_sampler import BatchWalkSampler
from hypergan.samplers.input_sampler import InputSampler
from hypergan.samplers.static_batch_sampler import StaticBatchSampler
from hypergan.samplers.y_sampler import YSampler
from pathlib import Path
from torch.autograd import Variable
from torch.autograd import grad as torch_grad
import hyperchamber as hc
import hypergan as hg
import inspect
import math
import numpy as np
import os
import re
import torch
import torch.nn as nn

class BaseGAN():
    def __init__(self, config=None, inputs=None):
        """ Initialized a new GAN."""
        self.steps = Variable(torch.zeros([1]))
        self.inputs = inputs
        self.inputs.gan = self
        self.components = {}
        self.destroy = False

        if config == None:
            config = hg.Configuration.default()

        self.config = config
        self._metrics = {}
        self.create()

    def add_metric(self, name, value):
        """adds metric to monitor during training
            name:string
            value:Tensor
        """
        self._metrics[name] = value
        return self._metrics

    def parameters(self):
        for param in self.g_parameters():
            yield param
        for param in self.d_parameters():
            yield param
        yield self.steps

    def g_parameters(self):
        print("Warning: BaseGAN.g_parameters() called directly.  Please override")

    def d_parameters(self):
        print("Warning: BaseGAN.d_parameters() called directly.  Please override")

    def metrics(self):
        """returns a metric : tensor hash"""
        return self._metrics

    def batch_size(self):
        return self.inputs.batch_size()

    def channels(self):
        return self.inputs.channels()

    def width(self):
        return self.inputs.width()

    def height(self):
        return self.inputs.height()

    def output_shape(self):
        return [self.width(), self.height(), self.channels()]

    def get_config_value(self, symbol):
        if symbol in self.config:
            config = hc.Config(hc.lookup_functions(self.config[symbol]))
            return config
        return None

    def add_component(self, name, component):
        index = 0
        while(name+str(index) in self.components):
            index+=1
        self.components[name+str(index)] = component

    def create_component(self, name, *args, **kw_args):
        print("Creating component:", name)
        defn = self.config[name]
        if defn == None:
            print("No definition found for " + name)
            return None
        if defn['class'] == None:
            raise ValidationException("Component definition is missing '" + name + "'")
        klass = GANComponent.lookup_function(None, defn['class'])
        gan_component = klass(self, defn, *args, **kw_args)
        if(isinstance(gan_component, nn.Module)):
            gan_component = gan_component.cuda()
        self.add_component(name, gan_component)
        return gan_component

    def create(self):
        print("Warning: BaseGAN.create() called directly.  Please override")

    def discriminator_fake_inputs(self, discriminator_index=0):
        """
            Fake inputs to the discriminator, should be cached
        """
        []

    def discriminator_real_inputs(self, discriminator_index=0):
        """
            Real inputs to the discriminator, should be cached
        """
        []

    def forward_discriminator(self, inputs, discriminator_index=0):
        """
            Runs a forward pass through the discriminator and returns the discriminator output
        """
        print("Warning: BaseGAN.forward_discriminator() called directly.  Please override")
        return None

    def forward_loss(self):
        """
            Runs a forward pass through the GAN and returns (d_loss, g_loss)
        """
        d_real, d_fake = self.forward_pass()
        return self.loss.forward(d_real, d_fake)

    def forward_pass(self):
        """
            Runs a forward pass through the GAN and returns (d_real, d_fake)
        """
        print("Warning: BaseGAN.forward_pass() called directly.  Please override")
        return None, None

    def step(self, feed_dict={}):
        self.steps += 1
        self._metrics = {}
        return self._step(feed_dict)

    def _step(self, feed_dict={}):
        if self.trainer == None:
            raise ValidationException("gan.trainer is missing.  Cannot train.")
        return self.trainer.step(feed_dict)

    def save(self, save_file):
        print("Saving..." + str(len(self.components)))
        full_path = os.path.expanduser(os.path.dirname(save_file))
        os.makedirs(full_path, exist_ok=True)
        for name, component in self.components.items():
            self._save(full_path, name, component)

    def _save(self, full_path, name, component):
        path = full_path + "/"+name+".save"
        print("Saving " + path)
        print(component.state_dict().keys())
        torch.save(component.state_dict(), path)

    def load(self, save_file):
        print("Loading..." + str(len(self.components)))
        success = True
        full_path = os.path.expanduser(os.path.dirname(save_file))
        for name, component in self.components.items():
            if not self._load(full_path, name, component):
                print("Error loading", name)
                success = False
        return success

    def _load(self, full_path, name, component):
        path = full_path + "/"+name+".save"
        if Path(path).is_file():
            print("Loading " + path)
            try:
                state_dict = torch.load(path)
                print('state_dict', state_dict.keys())
                component.load_state_dict(state_dict)
                return True
            except:
                print("Warning: Could not load component " + name)
                return False
        else:
            print("Could not load " + path)
            return False


    def parse_args(self, strs):
        options = hc.Config({})
        args = []
        for x in strs:
            if '=' in x:
                lhs, rhs = x.split('=')
                options[lhs]=rhs
            else:
                args.append(x)
        return args, options

    def exit(self):
        self.destroy = True

    def build(self, input_nodes=None, output_nodes=None):
        pass

    def get_registered_samplers(self=None):
        return {
                'static_batch': StaticBatchSampler,
                'input': InputSampler,
                #'progressive': ProgressiveSampler,
                #'random_walk': RandomWalkSampler,
                #'alphagan_random_walk': AlphaganRandomWalkSampler,
                #'style_walk': StyleWalkSampler,
                'batch_walk': BatchWalkSampler,
                'batch': BatchSampler,
                #'grid': GridSampler,
                #'sorted': SortedSampler,
                #'gang': GangSampler,
                #'began': BeganSampler,
                #'autoencode': AutoencodeSampler,
                #'debug': DebugSampler,
                'y': YSampler,
                #'segment': SegmentSampler,
                'aligned': AlignedSampler
            }

    def g_parameters(self):
        for component in self.generator_components():
            for param in component.parameters():
                yield param

    def d_parameters(self):
        for component in self.discriminator_components():
            for param in component.parameters():
                yield param

    def discriminator_components(self):
        print("Warning: BaseGAN.discriminator_components() called directly.  Please override")

    def generator_components(self):
        print("Warning: BaseGAN.generator_components() called directly.  Please override")

    def train_hooks(self):
        result = []
        for component in self.gan.components:
            if hasattr(component, "train_hooks"):
                result += component.train_hooks
        return result

    def sampler_for(self, name, default=StaticBatchSampler):
        samplers = self.get_registered_samplers()
        self.selected_sampler = name
        if name in samplers:
            return samplers[name]
        else:
            print("[hypergan] No sampler found for ", name, ".  Defaulting to", default)
            return default

    def regularize_adversarial_norm(self):
        raise ValidationException("Not implemented")

    def set_generator_trainable(self, flag):
        for c in self.generator_components():
            c.set_trainable(flag)

    def set_discriminator_trainable(self, flag):
        for c in self.discriminator_components():
            c.set_trainable(flag)



import hyperchamber as hc
import torch

from hypergan.losses.base_loss import BaseLoss

class SpecialDLoss(BaseLoss):
    def __init__(self, gan, config):
        super(BaseLoss, self).__init__(gan, config)
        self.special_discriminator = gan.create_component("special_discriminator")
        self.special_discriminator2 = gan.create_component("special_discriminator")

    def _forward(self, d_real_orig, d_fake_orig):
        criterion = torch.nn.BCEWithLogitsLoss()

        # a1 || d(x)
        # a0 || d(g)
        d_real = -self.special_discriminator(torch.rand_like(d_real_orig))
        d_real += self.special_discriminator(d_real_orig)
        d_real -= self.special_discriminator2(torch.randn_like(d_fake_orig))
        d_real += self.special_discriminator2(d_fake_orig)

        d_fake = -self.special_discriminator(torch.rand_like(d_fake_orig))
        d_fake += self.special_discriminator(d_fake_orig)
        d_fake -= self.special_discriminator2(torch.randn_like(d_real_orig))
        d_fake += self.special_discriminator2(d_real_orig)

        g_loss = -d_fake + d_real
        d_loss = -d_real + d_fake

        return [d_loss, g_loss]



#From https://gist.github.com/EndingCredits/b5f35e84df10d46cfa716178d9c862a3
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import state_ops
from tensorflow.python.framework import ops
from tensorflow.python.training import optimizer
import hyperchamber as hc
import numpy as np
import inspect
from operator import itemgetter
from hypergan.train_hooks.base_train_hook import BaseTrainHook

class ZeroPenaltyTrainHook(BaseTrainHook):
  def __init__(self, gan=None, config=None, trainer=None, name="GradientPenaltyTrainHook"):
    super().__init__(config=config, gan=gan, trainer=trainer, name=name)
    gan_inputs = self.gan.inputs.x

    self._lambda = self.gan.configurable_param(config['lambda'] or 1)

    if self.config.component:
        v = getattr(gan, self.config.component)
    else:
        v = gan.discriminator
    if self.config.layer is not None:
        layer = v.layer(self.config.layer)
    else:
        layer = v.sample
    self.loss = self._lambda * tf.reduce_sum(tf.square(layer))
    gan.add_metric('zero', self.loss)

  def losses(self):
    return [self.loss, self.loss]



from hypergan.samplers.base_sampler import BaseSampler
from hypergan.gan_component import ValidationException, GANComponent

import numpy as np
import time

class AlignedSampler(BaseSampler):
    def __init__(self, gan, samples_per_row=8):
        BaseSampler.__init__(self, gan, samples_per_row)
        self.inputs = self.gan.inputs.next()

    def compatible_with(gan):
        if hasattr(gan, 'encoder'):
            return True
        return False

    def _sample(self):
        #self.inputs = self.gan.inputs.next(0).clone().detach()
        g = self.gan.generator.forward(self.inputs)
        return [
            ('input', self.inputs),
            ('generator', g)
        ]




import hyperchamber as hc
import torch
import torch.nn as nn
import numpy as np

import torch.nn.functional as F
from hypergan.losses.base_loss import BaseLoss

class CategoricalLoss(nn.Module):
    def __init__(self, atoms=51, v_max=1.0, v_min=-1.0):
        super(CategoricalLoss, self).__init__()

        self.atoms = atoms
        self.v_max = v_max
        self.v_min = v_min
        self.supports = torch.linspace(v_min, v_max, atoms).view(1, 1, atoms).cuda() # RL: [bs, #action, #quantiles]
        self.delta = (v_max - v_min) / (atoms - 1)

    def forward(self, anchor, feature, skewness=0.0):
        batch_size = feature.shape[0]
        skew = torch.zeros((batch_size, self.atoms)).cuda().fill_(skewness)

        # experiment to adjust KL divergence between positive/negative anchors
        Tz = skew + self.supports.view(1, -1) * torch.ones((batch_size, 1)).to(torch.float).cuda().view(-1, 1)
        Tz = Tz.clamp(self.v_min, self.v_max)
        b = (Tz - self.v_min) / self.delta
        l = b.floor().to(torch.int64)
        u = b.ceil().to(torch.int64)
        l[(u > 0) * (l == u)] -= 1
        u[(l < (self.atoms - 1)) * (l == u)] += 1
        offset = torch.linspace(0, (batch_size - 1) * self.atoms, batch_size).to(torch.int64).unsqueeze(dim=1).expand(batch_size, self.atoms).cuda()
        skewed_anchor = torch.zeros(batch_size, self.atoms).cuda()
        skewed_anchor.view(-1).index_add_(0, (l + offset).view(-1), (anchor * (u.float() - b)).view(-1))  
        skewed_anchor.view(-1).index_add_(0, (u + offset).view(-1), (anchor * (b - l.float())).view(-1))  

        loss = -(skewed_anchor * (feature + 1e-16).log()).sum(-1).mean()

        return loss

class RealnessLoss(BaseLoss):
    """https://arxiv.org/pdf/2002.05512v1.pdf"""
    def __init__(self, gan, config):
        super(RealnessLoss, self).__init__(gan, config)

    def required(self):
        return "skew".split()

    def _forward(self, d_real, d_fake):
        num_outcomes = d_real.shape[1]
        if not hasattr(self, 'anchor0'):
            gauss = np.random.normal(0, 0.3, 1000)
            count, bins = np.histogram(gauss, num_outcomes)
            self.anchor0 = count / num_outcomes

            unif = np.random.uniform(-1, 1, 1000)
            count, bins = np.histogram(unif, num_outcomes)
            self.anchor1 = count / num_outcomes
            self.anchor_real = torch.zeros((self.gan.batch_size(), num_outcomes), dtype=torch.float).cuda() + torch.tensor(self.anchor1, dtype=torch.float).cuda()
            self.anchor_fake = torch.zeros((self.gan.batch_size(), num_outcomes), dtype=torch.float).cuda() + torch.tensor(self.anchor0, dtype=torch.float).cuda()
            self.Triplet_Loss = CategoricalLoss(num_outcomes)
        feat_real = d_real.log_softmax(1).exp()
        feat_fake = d_fake.log_softmax(1).exp()
        d_loss = self.Triplet_Loss(self.anchor_real, feat_real, skewness=self.config.skew[1]) + \
                 self.Triplet_Loss(self.anchor_fake, feat_fake, skewness=self.config.skew[0])

        g_loss = -self.Triplet_Loss(self.anchor_fake, feat_fake, skewness=self.config.skew[0])
        g_loss += self.Triplet_Loss(self.anchor_real, feat_fake, skewness=self.config.skew[0])

        return [d_loss, g_loss]




import torch.nn as nn
import torch
from torch.distributions import uniform

class ConcatNoise(nn.Module):
    def __init__(self):
        super(ConcatNoise, self).__init__()
        self.z = uniform.Uniform(torch.Tensor([-1.0]),torch.Tensor([1.0]))
    def forward(self, x):
        noise = self.z.sample(x.shape).cuda()
        cat = torch.cat([x, noise.view(*x.shape)], 1)
        return cat



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import state_ops
from tensorflow.python.framework import ops
from tensorflow.python.training import optimizer
import hyperchamber as hc
import inspect

class JROptimizer(optimizer.Optimizer):
  """https://arxiv.org/pdf/1806.09235.pdf"""
  def __init__(self, learning_rate=0.001, p=0.1, gan=None, config=None, use_locking=False, name="SOSOptimizer", optimizer=None, alpha=1):
    super().__init__(use_locking, name)
    self._alpha = alpha
    self.gan = gan
    self.config = config
    self._lr_t = learning_rate

    self.optimizer = self.gan.create_optimizer(optimizer)
 
  def _prepare(self):
    super()._prepare()
    self.optimizer._prepare()

  def _create_slots(self, var_list):
    super()._create_slots(var_list)
    self.optimizer._create_slots(var_list)

  def _apply_dense(self, grad, var):
    return self.optimizer._apply_dense(grad, var)

  def finite_differences(self, grads_and_vars, global_step, name, d_vars, g_vars, d_grads, g_grads):
    all_vars = [ v for _,v in grads_and_vars]
    all_grads = [ g for g, _ in grads_and_vars ]
    d_grads = all_grads[:len(d_vars)]
    g_grads = all_grads[len(d_vars):]
    d_vars = []
    g_vars = []
    for grad,var in grads_and_vars:
        if var in self.gan.d_vars():
            d_vars += [var]
        elif var in self.gan.g_vars():
            g_vars += [var]
        else:
            raise("Couldn't find var in g_vars or d_vars")

    with ops.init_scope():
        [self._zeros_slot(v, "orig", self._name) for _,v in grads_and_vars]
        slots_list = []
        if self.config.include_slots:
            for name in self.optimizer.get_slot_names():
                for var in self.optimizer.variables():
                    slots_list.append(self.optimizer._zeros_slot(var, "orig", "orig"))

    v1 = [self.get_slot(v, "orig") for _,v in grads_and_vars]
    slots_list = []
    slots_vars = []

    restored_vars = all_vars + slots_vars
    tmp_vars = v1 + slots_list

    e1 = 0.0001
    e2 = 0.0001

    #gamma12
    save = tf.group(*[tf.assign(w, v) for w,v in zip(tmp_vars, restored_vars)]) # store variables
    restore = tf.group(*[tf.assign(w, v) for w,v in zip(restored_vars, tmp_vars)]) # store variables

    def curl():
        grads = tf.gradients(self.gan.trainer.d_loss, d_vars) + tf.gradients(self.gan.trainer.g_loss, g_vars)
        op3 = tf.group(*[tf.assign_sub(v, self._lr_t*grad) for grad,v in zip(grads, all_vars)])
        with tf.get_default_graph().control_dependencies([op3]):
            def curlcombine(g1,g2):
                stepsize = self._lr_t
                return g1-(g2-g1)/stepsize
            new_grads = tf.gradients(self.gan.trainer.d_loss, d_vars) + tf.gradients(self.gan.trainer.g_loss, g_vars)
            g3s = [curlcombine(g1,g2) for g1,g2 in zip(grads,new_grads)]
            return g3s
 
    #gamma12
    with tf.get_default_graph().control_dependencies([save]):
        #opboth = self.optimizer.apply_gradients(grads_and_vars, global_step=global_step, name=name)
        #opdp = self.optimizer.apply_gradients(grads_and_vars[:len(d_vars)], global_step=global_step, name=name)
        #opgp = self.optimizer.apply_gradients(grads_and_vars[len(d_vars):], global_step=global_step, name=name)
        opboth = tf.group(*[tf.assign_sub(w, self._lr_t * v) for w,v in zip(all_vars, all_grads)]) # store variables
        opd = tf.group(*[tf.assign_sub(w, self._lr_t * v) for w,v in zip(d_vars, d_grads)]) # store variables
        opg = tf.group(*[tf.assign_sub(w, self._lr_t * v) for w,v in zip(g_vars, g_grads)]) # store variables
        with tf.get_default_graph().control_dependencies([opboth]):
            gboth = curl()#tf.gradients(self.gan.trainer.d_loss, d_vars) + tf.gradients(self.gan.trainer.g_loss, g_vars)
            with tf.get_default_graph().control_dependencies([restore]):
                with tf.get_default_graph().control_dependencies([opd]):
                    #new_d_grads = [tf.zeros_like(_d) for _d in d_vars]+tf.gradients(self.gan.trainer.g_loss, g_vars)
                    new_d_grads = curl()#tf.gradients(self.gan.trainer.d_loss, d_vars) + tf.gradients(self.gan.trainer.g_loss, g_vars)
                    with tf.get_default_graph().control_dependencies([restore]):
                        with tf.get_default_graph().control_dependencies([opg]):
                            #new_g_grads = tf.gradients(self.gan.trainer.d_loss, d_vars) + [tf.zeros_like(_g) for _g in g_vars]
                            new_g_grads = curl()#tf.gradients(self.gan.trainer.d_loss, d_vars) + tf.gradients(self.gan.trainer.g_loss, g_vars)
                            with tf.get_default_graph().control_dependencies([restore]):
                                new_grads = []
                                for _gboth, _gd, _gg, _g in zip(gboth,new_d_grads,new_g_grads,d_grads):
                                    det = tf.square(_gboth)-(_gg*_gd)+1e-8
                                    h_1 = 1.0/det * (2*_gboth - _gd - _gg)
                                    if self.config.hessian:
                                        #v = (g(x + hjej)-g(x)))/(2hj) + \
                                        #    (g(x + hiei)-g(x))/(2hi)
                                        a = (_gboth - _g) / self._lr_t # d2f/dx2i
                                        c = (_gboth - _g) / self._lr_t # d2f/dx2j
                                        b = (_gg - _g) / (2*self._lr_t)+(_gd-_g)/(2*self._lr_t) # d2f/dx1dx2
                                        d = b # d2f/dx2dx1
                                        det = a*d-b*c+1e-8
                                        #h_1 = 1.0/det * (b+d-a-c)
                                        h_1_a = d/det
                                        h_1_b = -b/det
                                        h_1_c = -c/det
                                        h_1_d = a/det

                                        h_1 = h_1_a*h_1_d-h_1_b*h_1_c
                                    new_grads.append( _g*h_1 )

                                for _gboth, _gd, _gg, _g in zip(gboth[len(d_vars):],new_d_grads[len(d_vars):],new_g_grads[len(d_vars):],g_grads):
                                    det = tf.square(_gboth)-(_gg*_gd)+1e-8
                                    h_1 = 1.0/det * (2*_gboth - _gd - _gg)
                                    if self.config.hessian:
                                        #v = (g(x + hjej)-g(x)))/(2hj) + \
                                        #    (g(x + hiei)-g(x))/(2hi)
                                        a = (_gboth - _g) / self._lr_t # d2f/dx2i
                                        c = (_gboth - _g) / self._lr_t # d2f/dx2j
                                        b = (_gg - _g) / (2*self._lr_t)+(_gd-_g)/(2*self._lr_t) # d2f/dx1dx2
                                        d = b # d2f/dx2dx1
                                        det = a*d-b*c+1e-8
                                        #h_1 = 1.0/det * (b+d-a-c)
                                        h_1_a = d/det
                                        h_1_b = -b/det
                                        h_1_c = -c/det
                                        h_1_d = a/det
                                        h_1 = h_1_a*h_1_d-h_1_b*h_1_c
                                    new_grads.append( _g*h_1 )

                                new_grads_and_vars = list(zip(new_grads, all_vars)).copy()
                                return self.optimizer.apply_gradients(new_grads_and_vars, global_step=global_step, name=name)

  def apply_gradients(self, grads_and_vars, global_step=None, name=None):
    all_vars = [ v for _,v in grads_and_vars]
    d_vars = []
    g_vars = []
    all_grads = [ g for g, _ in grads_and_vars ]
    for grad,var in grads_and_vars:
        if var in self.gan.d_vars():
            d_vars += [var]
        elif var in self.gan.g_vars():
            g_vars += [var]
        else:
            raise("Couldn't find var in g_vars or d_vars")

    with ops.init_scope():
        self.optimizer._create_slots([v for g,v in grads_and_vars])
    self._prepare()

    d_grads = all_grads[:len(d_vars)]
    g_grads = all_grads[len(d_vars):]
    if self.config.finite_differences:
        return self.finite_differences(grads_and_vars, global_step, name, d_vars, g_vars, d_grads, g_grads)
    dc_grads = sum([tf.reduce_sum(tf.square(d)) for d in d_grads])
    gc_grads = sum([tf.reduce_sum(tf.square(g)) for g in g_grads])
    gamma12 = tf.gradients(gc_grads, d_vars) + [tf.zeros_like(g) for g in g_vars]
    gamma21 = [tf.zeros_like(d) for d in d_vars] + tf.gradients(dc_grads, g_vars)

    gamma12 = [ tf.zeros_like(ddg) if _dg is None else _dg for ddg, _dg in zip(all_vars, gamma12) ]
    gamma21 = [ tf.zeros_like(ddg) if _dg is None else _dg for ddg, _dg in zip(all_vars, gamma21) ]
    __gamma12 = [ tf.reduce_sum(_gamma12) for _gamma12 in gamma12 ]
    __gamma21 = [ tf.reduce_sum(_gamma21) for _gamma21 in gamma21 ]
    #gamma12_metric = self.gan.ops.squash(sum(gamma12))
    gamma12_metric = self.gan.ops.squash(sum(__gamma12))
    self.gan.add_metric('gamma12', gamma12_metric)
    gamma21_metric = self.gan.ops.squash(sum(__gamma21))
    self.gan.add_metric('gamma21', gamma21_metric)
   
    new_grads = []
    for _gamma12, _gamma21, _grads in zip(gamma12, gamma21, all_grads):
        Eo = _grads - \
             0.5*self._alpha*_gamma21 +\
             0.5*self._alpha*_gamma12
        new_grads += [ Eo ]

    new_grads_and_vars = list(zip(new_grads, all_vars)).copy()

    return self.optimizer.apply_gradients(new_grads_and_vars, global_step=global_step, name=name)
  
  def _apply_sparse(self, grad, var):
    raise NotImplementedError("Sparse gradient updates are not supported.")



import numpy as np
import hyperchamber as hc
import torch

from hypergan.losses.base_loss import BaseLoss

TINY=1e-8
class FDivergenceLoss(BaseLoss):
    def __init__(self, gan, config):
        super(FDivergenceLoss, self).__init__(gan, config)
        self.tanh = torch.nn.Tanh()

    def _forward(self, d_real, d_fake):
        gan = self.gan
        config = self.config

        gfx = None
        gfg = None

        pi = config.pi or 2

        g_loss_type = config.g_loss_type or config.type or 'gan'
        d_loss_type = config.type or 'gan'

        alpha = config.alpha or 0.5

        if d_loss_type == 'kl':
            bounded_x = torch.clamp(d_real, max=np.exp(9.))
            bounded_g = torch.clamp(d_fake, max=10.)
            gfx = bounded_x
            gfg = bounded_g
        elif d_loss_type == 'js':
            gfx = np.log(2) - (1+(-d_real).exp()).log()
            gfg = np.log(2) - (1+(-d_fake).exp()).log()
        elif d_loss_type == 'js_weighted':
            gfx = -pi*np.log(pi) - (1+(-d_real).exp()).log()
            gfg = -pi*np.log(pi) - (1+(-d_fake).exp()).log()
        elif d_loss_type == 'gan':
            gfx = -(1+(-d_real).exp()).log()
            gfg = -(1+(-d_fake).exp()).log()
        elif d_loss_type == 'reverse_kl':
            gfx = -d_real.log()
            gfg = -d_fake.log()
        elif d_loss_type == 'pearson' or d_loss_type == 'jeffrey' or d_loss_type == 'alpha2':
            gfx = d_real
            gfg = d_fake
        elif d_loss_type == 'squared_hellinger':
            gfx = 1-(-d_real).exp()
            gfg = 1-(-d_fake).exp()
        elif d_loss_type == 'neyman':
            gfx = 1-(-d_real).exp()
            gfx = torch.clamp(gfx, max=1.9)
            gfg = 1-(-d_fake).exp()

        elif d_loss_type == 'total_variation':
            gfx = 0.5*self.tanh(d_real)
            gfg = 0.5*self.tanh(d_fake)

        elif d_loss_type == 'alpha1':
            gfx = 1./(1-alpha) - (1+(-d_real).exp()).log()
            gfg = 1./(1-alpha) - (1+(-d_fake).exp()).log()

        else:
            raise "Unknown type " + d_loss_type

        conjugate = None

        if d_loss_type == 'kl':
            conjugate = (gfg-1).exp()
        elif d_loss_type == 'js':
            bounded = torch.clamp(gfg, max=np.log(2.)-TINY)
            conjugate = -(2-(bounded).exp()).log()
        elif d_loss_type == 'js_weighted':
            c = -pi*np.log(pi)-TINY
            bounded = gfg
            conjugate = (1-pi)*((1-pi)/((1-pi)*(bounded/pi).exp())).log()
        elif d_loss_type == 'gan':
            conjugate = -(1-(gfg).exp()).log()
        elif d_loss_type == 'reverse_kl':
            conjugate = -1-(-gfg).log()
        elif d_loss_type == 'pearson':
            conjugate = 0.25 * (gfg**2)+gfg
        elif d_loss_type == 'neyman':
            conjugate = 2 - 2 * torch.sqrt(self.relu(1-gfg)+1e-2)
        elif d_loss_type == 'squared_hellinger':
            conjugate = gfg/(1.-gfg)
        elif d_loss_type == 'jeffrey':
            raise "jeffrey conjugate not implemented"

        elif d_loss_type == 'alpha2' or d_loss_type == 'alpha1':
            bounded = gfg
            bounded = 1./alpha * (bounded * ( alpha - 1) + 1)
            conjugate = (bounded ** ( alpha/(alpha - 1.))) - 1. / alpha

        elif d_loss_type == 'total_variation':
            conjugate = gfg
        else:
            raise "Unknown type " + d_loss_type

        gf_threshold  = None # f' in the paper

        if d_loss_type == 'kl':
            gf_threshold = 1
        elif d_loss_type == 'js':
            gf_threshold = 0
        elif d_loss_type == 'gan':
            gf_threshold = -np.log(2)
        elif d_loss_type == 'reverse_kl':
            gf_threshold = -1
        elif d_loss_type == 'pearson':
            gf_threshold = 0
        elif d_loss_type == 'squared_hellinger':
            gf_threshold = 0

        self.gf_threshold=gf_threshold

        d_loss = -gfx+conjugate
        g_loss = -conjugate

        if g_loss_type == 'gan':
            g_loss = -conjugate
        elif g_loss_type == 'total_variation':
            # The inverse of derivative(1/2*x - 1)) = 0.5
            # so we use the -conjugate for now
            g_loss = -conjugate
        elif g_loss_type == 'js':
            # https://www.wolframalpha.com/input/?i=inverse+of+derivative(-(u%2B1)*log((1%2Bu)%2F2)%2Bu*log(u))
            g_loss = -d_fake.exp()
        elif g_loss_type == 'js_weighted':
            # https://www.wolframalpha.com/input/?i=inverse+of+derivative(-(u%2B1)*log((1%2Bu)%2F2)%2Bu*log(u))
            p = pi
            u = d_fake
            exp_bounded = p/u
            exp_bounded = torch.clamp(exp_bounded, max=4.)
            inner = (-4.*u*(exp_bounded).exp() +np.exp(2.)*(u**2)-2.*np.exp(2.)*u+np.exp(2.))/(u**2)
            inner = self.relu(inner)
            u = torch.clamp(u, max=0.1)
            sqrt = torch.sqrt(inner+1e-2) / (2*np.exp(1))
            g_loss = (1.-u)/(2.*u)# + sqrt
        elif g_loss_type == 'pearson':
            g_loss = -(d_fake-2.0)/2.0
        elif g_loss_type == 'neyman':
            g_loss = 1./torch.sqrt(1-d_fake) # does not work, causes 'nan'
        elif g_loss_type == 'squared_hellinger':
            g_loss = -1.0/(((d_fake-1)**2)+1e-2)
        elif g_loss_type == 'reverse_kl':
            g_loss = -d_fake
        elif g_loss_type == 'kl':
            g_loss = -gfg * gfg.exp()
        elif g_loss_type == 'alpha1': 
            a = alpha
            bounded = d_fake
            g_loss = (1.0/(a*(a-1))) * ((a*bounded).exp() - 1 - a*((bounded).exp() - 1))
        elif g_loss_type == 'alpha2':
            a = alpha
            bounded = torch.clamp(d_fake, max=4.)
            g_loss = -(1.0/(a*(a-1))) * ((a*bounded).exp() - 1 - a*((bounded).exp() - 1))
        else:
            raise "Unknown g_loss_type " + g_loss_type

        if self.config.regularizer:
            g_loss += self.g_regularizer(gfg, gfx)

        return [d_loss, g_loss]

    def g_regularizer(self, gfg, gfx):
        regularizer = None
        config = self.config
        pi = config.pi or 2
        alpha = config.alpha or 0.5

        ddfc = 0

        if config.regularizer == 'kl':
            # https://www.wolframalpha.com/input/?i=derivative(derivative(exp(t-1)))
            bounded = torch.clamp(gfg, max=4.)
            ddfc = (bounded - 1).exp()
        elif config.regularizer == 'js':
            # https://www.wolframalpha.com/input/?i=derivative(derivative(-log(2-exp(t))))
            ddfc = -(2*(gfg).exp()) / ((2-gfg.exp())**2+1e-2)
        elif config.regularizer == 'js_weighted':
            # https://www.wolframalpha.com/input/?i=derivative(derivative((1-C)*log(((1-C)%2F(1-C*exp(x%2FC)))))
            ddfc = -((pi-1)*(gfg/pi).exp())/(pi*((pi*(gfg/pi).exp()-1)**2))
        elif config.regularizer == 'gan':
            # https://www.wolframalpha.com/input/?i=derivative(derivative(-log(1-exp(t))))
            ddfc = (2*gfg.exp()) / ((1-gfg.exp())**2+1e-2)
        elif config.regularizer == 'reverse_kl':
            # https://www.wolframalpha.com/input/?i=derivative(derivative(-1-log(-x)))
            ddfc = 1.0/(gfg**2)
        elif config.regularizer == 'pearson':
            # https://www.wolframalpha.com/input/?i=derivative(derivative(0.25*x*x+%2B+x))
            ddfc = 0.5
        elif config.regularizer == 'jeffrey':
            raise "jeffrey regularizer not implemented"
        elif config.regularizer == 'squared_hellinger': 
            # https://www.wolframalpha.com/input/?i=derivative(derivative(t%2F(1-t)))
            ddfc = 2 / (((gfg - 1) ** 3)+1e-2)
            #ddfc = 0
        elif config.regularizer == 'neyman':
            # https://www.wolframalpha.com/input/?i=derivative(derivative(2-2*sqrt(1-t)))
            ddfc = 1.0/(2*((1-gfg ** 3/2.)))
        elif config.regularizer == 'total_variation':
            # https://www.wolframalpha.com/input/?i=derivative(derivative(t))
            ddfc = 0
        elif config.regularizer == 'alpha1' or config.regularizer == 'alpha2':
            # https://www.wolframalpha.com/input/?i=derivative(derivative(1%2FC*(x*(C-1)%2B1)%5E(C%2FC-1)-1%2FC))
            ddfc = -(((alpha - 1)*gfg+1) ** (1/(alpha-1)-1))
        regularizer = ddfc * torch.norm(gfg, 2, 0) * (config.regularizer_lambda or 1)
        self.add_metric('fgan_regularizer', regularizer.mean())
        return regularizer 
        
    def d_regularizers(self):
        return []



import torch.nn as nn
import hypergan as hg

class Residual(hg.Layer):
    """
        ---
        description: 'layer residual for configurable component'
        ---

        # residual layer

        `residual` adds one or more residual blocks https://paperswithcode.com/method/residual-block

        ## optional arguments

        The number of residual blocks to add

        ## input size

        Any 4-d tensor

        ## output size

        Same as input size

        ## syntax

        ```json
          "residual COUNT"
        ```

        ## examples

        ```json
          "residual 3"
        ```
    """


    def __init__(self, component, args, options):
        super(Residual, self).__init__(component, args, options)
        self.size = component.current_size
        layers = []
        for i in range(options.count or 3):
            layers += [nn.Conv2d(self.size.channels, self.size.channels, 3, 1, padding = (1, 1))]
            layers += [nn.ReLU()]
            layers += [nn.Conv2d(self.size.channels, self.size.channels, 3, 1, padding = (1, 1))]
            layers += [nn.ReLU()]

        self.residual = nn.Sequential(*layers)

    def output_size(self):
        return self.size

    def forward(self, input, context):
        residual = self.residual(input)
        return input + residual



#https://arxiv.org/pdf/1811.07296.pdf
import hyperchamber as hc
from functools import reduce

from hypergan.losses.base_loss import BaseLoss

class QPLoss(BaseLoss):
    """https://arxiv.org/abs/1811.07296"""
    def _forward(self, d_real, d_fake):
        gan = self.gan

        pq = d_real
        pp = d_fake

        lam = 10.0/(reduce(lambda x,y:x*y, gan.output_shape()))
        dist = (gan.generator.sample - self.gan.inputs.sample).abs().mean()

        dl = - d_real + d_fake
        d_norm = 10 * dist
        d_loss = ( dl + 0.5 * dl**2 / d_norm).mean()

        g_loss = d_real - d_fake

        return [d_loss, g_loss]



import torch.nn as nn
import torch
from hypergan.layer_shape import LayerShape
import hypergan as hg

class SegmentSoftmax(hg.Layer):
    """
        ---
        description: 'layer segment_softmax for configurable component'
        ---

        # segment_softmax layer

        `segment_softmax` is a custom layer that allows for masking multiple output channels.

        Suppose you have 30 channels and `segment_softmax 3`. First, the 30 channels split into 15/15.
        The first 15 will be used for softmax and multiplied against the second.
        Then each channel is softmaxed, multiplied, and summed.

        So 30 input channels with 3 output channels equate to 5 input channels for each output channel.

        ## input size

        Any 4-d tensor of the shape `[B, C, H, W]`

        ## output size

        [B, OUTPUT_CHANNELS, H, W]

        ## syntax

        ```json
          "segment_softmax OUTPUT_CHANNELS"
        ```

        ## examples

        At the end of the generator for RGB images:

        ```json
          "conv 30",
          "segment_softmax 3",
          "hardtanh"
        ```
    """


    def __init__(self, component, args, options):
        super(SegmentSoftmax, self).__init__(component, args, options)
        self.channels = args[0]
        self.dims = list(component.current_size.dims).copy()
        self.softmax = nn.Softmax(dim=2)

    def output_size(self):
        return LayerShape(*([self.channels]+self.dims[1:]))

    def forward(self, input, context):
        content, segment = torch.split(input, input.shape[1]//2, 1)
        net_in = content.view(content.shape[0], content.shape[1]//self.channels, self.channels, content.shape[2], content.shape[3])
        segment = segment.view(content.shape[0], content.shape[1]//self.channels, self.channels, content.shape[2], content.shape[3])
        selection = self.softmax(segment)
        rendered = (selection * net_in).sum(dim=1).view([input.shape[0]]+list(self.output_size().dims))
        return rendered



import math
import hypergan as hg
import torch
import torch.nn as nn

from hypergan.layer_shape import LayerShape

class LinearAttention(hg.Layer):
    def __init__(self, component, args, options):
        super(LinearAttention,self).__init__(component, args, options)
        input_size = component.current_size.size()
        output_size = args[0]

        self.heads = options.heads or 4
        self.query = nn.Linear(input_size, input_size).cuda()
        self.key = nn.Linear(input_size, input_size).cuda()
        self.value = nn.Linear(input_size, input_size).cuda()
        self.out = nn.Linear(input_size, output_size).cuda()
        self.eps = torch.tensor(1e-12).cuda()
        self.size = LayerShape(output_size)
        if 'w' in component.layer_output_sizes:
            self.style = nn.Linear(component.layer_output_sizes['w'].size(), output_size).cuda() 

        component.nn_init(self.query, options.initializer)
        component.nn_init(self.key, options.initializer)
        component.nn_init(self.value, options.initializer)
        component.nn_init(self.out, options.initializer)

    def elu_feature_map(self, x):
        return torch.nn.functional.elu(x) + 1

    def forward(self, x, context):
        queries = self.query(x)
        keys = self.key(x)
        values = self.value(x)
        if 'w' in context:
            style = self.style(context['w'])
        else:
            style = None
        N, L = queries.shape
        D = 1
        _, S = keys.shape
        H = self.heads
        Q = self.elu_feature_map(queries.view(N, L, H, -1))
        K = self.elu_feature_map(keys.view(N, L, H, -1))
        V = values.view(N, L, H, -1)
        KV = torch.einsum("nshd,nshm->nhmd", K, V)
        Z = 1/(torch.einsum("nlhd,nhd->nlh", Q, K.sum(dim=1))+self.eps)
        if 'w' in context:
            print("size", KV.shape, style.shape, K.shape, V.shape)
            V = torch.einsum("nlhd,nhmd,nlh->nlhm", Q, KV, Z)*style
        else:
            V = torch.einsum("nlhd,nhmd,nlh->nlhm", Q, KV, Z)

        return self.out(V.contiguous().view(x.shape))

    def output_size(self):
        return self.size



#From https://gist.github.com/EndingCredits/b5f35e84df10d46cfa716178d9c862a3
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import state_ops
from tensorflow.python.framework import ops
from tensorflow.python.training import optimizer
import hyperchamber as hc
import inspect

class DepthOptimizer(optimizer.Optimizer):
  """Steps multiple times and decays results"""
  def __init__(self, learning_rate=0.001, decay=0.9, gan=None, config=None, use_locking=False, name="depthOptimizer", optimizer=None, depth=3):
    super().__init__(use_locking, name)
    self._decay = decay
    self._depth = depth
    self.gan = gan
    self.config = config
    self.name = name
    self.optimizer = self.gan.create_optimizer(optimizer)

  def _prepare(self):
    super()._prepare()
    self.optimizer._prepare()

  def _create_slots(self, var_list):
    super()._create_slots(var_list)
    self.optimizer._create_slots(var_list)

  def apply_gradients(self, grads_and_vars, global_step=None, name=None):
    d_vars = []
    g_vars = []
    for grad,var in grads_and_vars:
        if var in self.gan.d_vars():
            d_vars += [var]
        elif var in self.gan.g_vars():
            g_vars += [var]
        else:
            raise Exception("Couldn't find var in g_vars or d_vars")

    if self.config.apply_on == "discriminator":
        depth_vars = d_vars
    else:
        depth_vars = d_vars + g_vars
    with ops.init_scope():
        [self._get_or_make_slot(v, v, "depth", self.name) for v in depth_vars]
        self.optimizer._create_slots([v for g,v in grads_and_vars])
        for name in self.optimizer.get_slot_names():
            for var in self.optimizer.variables():
                self._zeros_slot(var, "depth", self.name)

    self._prepare()
    depth_slots = [self.get_slot(v, "depth") for v in depth_vars]
    for name in self.optimizer.get_slot_names():
        for var in self.optimizer.variables():
            depth_vars += [var]
            depth_slots += [self._zeros_slot(var, "depth", self.name)]

    def calculate_depth(grads_and_vars_k,k=0):
        if(k == 0):
            return tf.group(*[tf.assign(v,nv) for v,nv in zip(depth_vars, depth_slots)])

        op2 = self.optimizer.apply_gradients(grads_and_vars_k, global_step=global_step, name=name)
        with tf.get_default_graph().control_dependencies([op2]):
            w_k_combined = [self._decay *w_k_1 + (1.-self._decay)*w_hat for w_hat, w_k_1 in zip(depth_slots, depth_vars)]
            op3 = tf.group(*[tf.assign(w, v) for w,v in zip(depth_slots, w_k_combined)]) # store variables
            with tf.get_default_graph().control_dependencies([op3]):
                d_loss, g_loss = self.gan.loss.sample
                d_grads = tf.gradients(d_loss, d_vars)
                g_grads = tf.gradients(g_loss, g_vars)
                grads_k_1 = d_grads + g_grads
                grads_and_vars_k_1 = list(zip(grads_k_1,depth_vars)).copy()
                return calculate_depth(grads_and_vars_k_1,k-1)

    op1 = tf.group(*[tf.assign(w, v) for w,v in zip(depth_slots, depth_vars)]) # store variables
    with tf.get_default_graph().control_dependencies([op1]):
        opd = calculate_depth(grads_and_vars, self._depth)
        with tf.get_default_graph().control_dependencies([opd]):
            return tf.no_op()

  
  def _apply_sparse(self, grad, var):
    raise NotImplementedError("Sparse gradient updates are not supported.")

  def _apply_dense(self, grad, var):
    raise NotImplementedError("_apply_dense not callable.")

  def variables(self):
      return self.optimizer.variables()



#From https://gist.github.com/EndingCredits/b5f35e84df10d46cfa716178d9c862a3
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import state_ops
from tensorflow.python.framework import ops
from tensorflow.python.training import optimizer
import hyperchamber as hc
import inspect

class ConsensusOptimizer(optimizer.Optimizer):
  def __init__(self, learning_rate=0.001, p=0.1, gan=None, config=None, use_locking=False, name="CurlOptimizer", optimizer=None, beta=1):
    super().__init__(use_locking, name)
    self._beta = beta
    self.gan = gan
    self.config = config
    self._lr_t = learning_rate
    self.optimizer = self.gan.create_optimizer(optimizer)
 
  def _prepare(self):
    super()._prepare()
    self.optimizer._prepare()

  def _create_slots(self, var_list):
    super()._create_slots(var_list)
    self.optimizer._create_slots(var_list)

  def _apply_dense(self, grad, var):
    return self.optimizer._apply_dense(grad, var)

  def apply_gradients(self, grads_and_vars, global_step=None, name=None):
    var_list = [ v for _,v in grads_and_vars]
    d_vars = []
    g_vars = []
    all_grads = [ g for g, _ in grads_and_vars ]
    for grad,var in grads_and_vars:
        if var in self.gan.d_vars():
            d_vars += [var]
        elif var in self.gan.g_vars():
            g_vars += [var]
        else:
            raise("Couldn't find var in g_vars or d_vars")

    with ops.init_scope():
        self.optimizer._create_slots([v for g,v in grads_and_vars])

    self._prepare()
    d_grads = all_grads[:len(d_vars)]
    if self.config.type == 'sga':
        Jgrads = tf.gradients(d_grads, d_vars, grad_ys=d_grads, stop_gradients=d_vars) + [tf.zeros_like(g) for g in g_vars]
    elif self.config.type == 'magnitude':
        consensus_reg = [tf.square(g) for g in d_grads if g is not None]
        Jgrads = tf.gradients(consensus_reg, d_vars) + [tf.zeros_like(g) for g in g_vars]
    else:
        consensus_reg = 0.5 * sum(
                tf.reduce_sum(tf.square(g)) for g in d_grads if g is not None
        )
        Jgrads = tf.gradients(consensus_reg, d_vars, stop_gradients=d_vars) + [tf.zeros_like(g) for g in g_vars]
    new_grads = [g+jg*self._beta if jg is not None else g for g,v,jg in zip(all_grads, var_list, Jgrads)]
    new_grads_and_vars = list(zip(new_grads, var_list)).copy()
    return self.optimizer.apply_gradients(new_grads_and_vars, global_step=global_step, name=name)

  
  def _apply_sparse(self, grad, var):
    raise NotImplementedError("Sparse gradient updates are not supported.")



import numpy as np
import hyperchamber as hc
import inspect

from hypergan.trainers.consensus_trainer import ConsensusTrainer
from hypergan.trainers.base_trainer import BaseTrainer

TINY = 1e-12

class KBeamTrainer(BaseTrainer):

    def _create(self):
        gan = self.gan
        config = self.config

        d_vars = gan.discriminator.d_variables
        g_vars = self.g_vars
        loss = self.loss
        trainers = []

        for l, d in zip(gan.loss.losses, d_vars):
            trainers += [ConsensusTrainer(self.gan, self.config, loss=l, d_vars=d, g_vars=g_vars)]

        self.trainers = trainers
        self.hist = [0 for i in range(len(self.trainers))]
        self.tfsummary_writer = tf.summary.FileWriter('./logs/sess.graph', tf.get_default_graph())
        tf.summary.scalar("zero", tf.reduce_mean(gan.loss.losses[0].sample))
        self.tfmerge_summary = tf.summary.merge_all()
        return None, None

    def required(self):
        return "trainer learn_rate".split()

    def _step(self, feed_dict):
        gan = self.gan
        sess = gan.session
        config = self.config
        loss = self.loss or gan.loss
        metrics = loss.metrics
        
        losses_g = [t.g_loss for t in self.trainers]
        losses_d = [t.d_loss for t in self.trainers]
        targets_t = losses_g+losses_d
        targets = sess.run(targets_t)
        l_g = targets[:len(targets)//2]
        l_d = targets[len(targets)//2:]

        if config.criteria == '<g':
            i=np.argmin([float(x) for x in l_g])
        elif config.criteria == '>d':
            i=np.argax([float(x) for x in l_d])
        elif config.criteria == '<d':
            i=np.argmin([float(x) for x in l_d])
        else:
            # default from paper
            i=np.argmax([float(x) for x in l_g])

        i=np.argmax([float(x) for x in l_g])
        self.hist[i]+=1
        optimizer = self.trainers[i].optimizer

        for t_t, t in zip(targets_t, targets):
            feed_dict[t_t] = t

        if config.alternating_trainer:
            _ = sess.run(self.trainers[i].g_optimizer, feed_dict)
            metric_values = sess.run([t.d_optimizer for j, t in enumerate(self.trainers)] + self.output_variables(metrics))[len(self.trainers):]
        elif config.train_all:
            metric_values = sess.run([t.optimizer for j, t in enumerate(self.trainers)] + self.output_variables(metrics), feed_dict)[len(self.trainers):]
        else:
            metric_values = sess.run([t.optimizer if j == i else t.d_optimizer for j, t in enumerate(self.trainers)] + self.output_variables(metrics), feed_dict)[len(self.trainers):]

        if self.current_step % 100 == 0:
            hist_output = "  " + "".join(["D"+str(i)+":"+str(v)+" "for i, v in enumerate(self.hist)])
            print(self.output_string(metrics) % tuple([self.current_step] + metric_values)+hist_output)
            self.hist = [0 for i in range(len(self.trainers))]




from subprocess import call

import glob

dirnames = glob.glob("samples/*")

for d in dirnames:
    images = glob.glob(d+"/*.png")

    print("")
    print("### "+d)
    images.sort()
    for image in images:
        print("")
        print("!["+image+"]("+image+")")





from subprocess import call

import glob

filenames = glob.glob("*.json")

cmd="rm -rf samples"
result = call(cmd, shell=True)

for f in filenames:
    config = f.split(".")[0]

    cmd="CUDA_VISIBLE_DEVICES=0 hypergan train /ml/datasets/faces/128x128/all --sample_every 8000 --sampler debug --format jpg --size 64x64x3 -b 8 -c ./"+config+" --resize --save_samples --steps 16001 --save_every 1000000"
    print(cmd)
    result = call(cmd, shell=True)

    cmd="mkdir samples/"+config
    result = call(cmd, shell=True)

    cmd="mv samples/*.png samples/"+config
    result = call(cmd, shell=True)





import torch.nn as nn
import torch

import hypergan as hg
import hyperchamber as hc
import pyparsing

from hypergan.gan_component import ValidationException

class Operation(hg.Layer):
    """
        ---
        description: Base class for operations
        ---

        This is a base class for many operations and not used directly.
    """
    def __init__(self, operation, component, args, options):
        super(Operation, self).__init__(component, args, options)
        self.operation = operation
        self.size = component.current_size
        self.layers, self.layer_names = self.build_layers(component, args, options)
        for i, (layer, layer_name) in enumerate(zip(self.layers, self.layer_names)):
            self.add_module('layer_'+str(i)+"_"+layer_name, layer)

    def build_layers(self, component, args, options):
        options = hc.Config(options)
        layers = []
        layer_names = []

        for arg in args:
            component.current_size = self.size
            if arg == 'self':
                layers.append(None)
                layer_names.append("self")
            elif arg == 'noise':
                layers.append(LearnedNoise())
                layer_names.append(None)
            elif arg in component.named_layers:
                layers.append(None)
                layer_names.append("layer "+arg)
            elif arg in component.gan.named_layers:
                layers.append(component.gan.named_layers[arg])
                layer_names.append(None)
            elif arg in component.context_shapes:
                layers.append(None)
                layer_names.append(arg)
            elif type(arg) == pyparsing.ParseResults and type(arg[0]) == hg.parser.Pattern:
                parsed = arg[0]
                parsed.parsed_options = hc.Config(parsed.options)
                layers.append(component.build_layer(parsed.layer_name, parsed.args, parsed.parsed_options))
                layer_names.append(parsed.layer_name)
            else:
                raise ValidationException("Could not parse operation layer '" + arg + "'")
        return layers, layer_names

    def output_size(self):
        return self.size

    def forward(self, input, context):
        output = None
        for layer, layer_name in zip(self.layers, self.layer_names):
            if layer_name == "self":
                layer_output = input
            elif isinstance(layer, hg.Layer):
                layer_output = layer(input, context)
            elif layer_name.split(" ")[0] == 'layer':
                layer_output = context[layer_name.split(" ")[1]]
            elif layer_name in context:
                layer_output = context[layer_name]
            else:
                layer_output = layer(input)
            if output is None:
                output = layer_output
            else:
                if self.operation == "+":
                    output = output + layer_output
                elif self.operation == "*":
                    output = output * layer_output
                elif self.operation == "cat":
                    output = torch.cat([output, layer_output], 1)
                else:
                    raise ValidationException("Unknown operation: "+ self.operation)
        return output



import hyperchamber as hc
import torch

from hypergan.losses.base_loss import BaseLoss

class LogisticLoss(BaseLoss):
    """ported from stylegan"""
    def __init__(self, gan, config):
        super(LogisticLoss, self).__init__(gan, config)
        self.softplus = torch.nn.Softplus(self.config.beta or 1, self.config.threshold or 20)

    def _forward(self, d_real, d_fake):
        d_loss = self.softplus(-d_real) + self.softplus(d_fake)
        g_loss = self.softplus(-d_fake)

        return [d_loss, g_loss]



import hypergan as hg
from hypergan.layer_shape import LayerShape
from . import Operation

class Add(Operation):
    """
        ---
        description: 'layer add for configurable component'
        ---

        # add layer

        Adds two or more layers together. Accepts nested layer definitions.

        ## input size

        Any number of matching tensors

        ## output size

        Same as input size

        ## syntax

        ```json
          "add [layer]*"
        ```

        ## examples

        ```json
          "add self (attention)"
        ```
    """
    def __init__(self, component, args, options):
        super(Add, self).__init__("+", component, args, options)




import torch.nn as nn
import hypergan as hg
from hypergan.layer_shape import LayerShape
from hypergan.modules.multi_head_attention import MultiHeadAttention as module_MultiHeadAttention

class MultiHeadAttention(hg.Layer):
    """
        ---
        description: 'layer multi_head_attention for configurable component'
        ---

        # multi_head_attention layer

        Adds an attention layer with multiple heads. Uses softmax. Ends in a linear

        ## required arguments

            `size` - output size

        ## optional arguments

            `heads` - Number of heads to use. Defaults to 4

        ## input size

        Any 2-d tensor

        ## output size

        First argument

        ## syntax

        ```json
          "ez_norm CHANNELS heads=HEADS"
        ```

        ## examples

        ```json
          "multi_head_attention 1024 heads=4"
        ```
    """
    def __init__(self, component, args, options):
        super(MultiHeadAttention, self).__init__(component, args, options)
        self.size = component.current_size.size()
        if len(args) > 0:
            self.size = args[0]
        self.layer = module_MultiHeadAttention(component.current_size.size(), self.size, heads=options.heads or 4)
        component.nn_init(self.layer.o, options.initializer)
        component.nn_init(self.layer.h, options.initializer)
        component.nn_init(self.layer.g, options.initializer)
        component.nn_init(self.layer.f, options.initializer)

    def output_size(self):
        return LayerShape(self.size)

    def forward(self, input, context):
        return self.layer.forward(input)





import importlib
import json
import numpy as np
import os
import sys
import time
import uuid
import copy

from hypergan.discriminators import *
from hypergan.distributions import *
from hypergan.generators import *
from hypergan.inputs import *
from hypergan.samplers import *
from hypergan.trainers import *

import hyperchamber as hc
from hyperchamber import Config
from hypergan.ops import TensorflowOps
import hypergan as hg

from hypergan.gan_component import ValidationException, GANComponent
from ..standard_gan import StandardGAN
from ..base_gan import BaseGAN

class AutoencoderGAN(StandardGAN):
    """ 
    """

    def required(self):
        return "generator".split()

    def create(self):
        config = self.config

        d2 = dict(config.discriminator)
        d2['class'] = self.ops.lookup("class:hypergan.discriminators.pyramid_discriminator.PyramidDiscriminator")
        self.encoder = self.create_component(d2)
        self.encoder.ops.describe("encoder")
        self.encoder.create(self.inputs.x)
        self.encoder.z = tf.zeros(0)
        self.trainer = self.create_component(config.trainer)

        StandardGAN.create(self)
        cycloss = tf.reduce_mean(tf.abs(self.inputs.x-self.generator.sample))
        cycloss_lambda = config.cycloss_lambda or 10
        self.loss.sample[1] *= config.g_lambda or 1
        self.loss.sample[1] += cycloss*cycloss_lambda
        self.trainer.create()

        self.initialize_variables()




from hypergan.samplers.base_sampler import BaseSampler

from hypergan.samplers.began_sampler import BeganSampler
from hypergan.samplers.batch_sampler import BatchSampler
from hypergan.samplers.static_batch_sampler import StaticBatchSampler
from hypergan.samplers.random_walk_sampler import RandomWalkSampler
from hypergan.samplers.segment_sampler import SegmentSampler
from hypergan.trainers.gang_trainer import GangTrainer
import numpy as np
import hypergan as hg

class GangSampler(BaseSampler):
    def __init__(self, gan, samples_per_row=8):
        BaseSampler.__init__(self, gan, samples_per_row=samples_per_row)
        self.xs = None
        self.samples = 1

    def compatible_with(gan):
        return isinstance(gan.trainer, GangTrainer)

    def sample(self, path, sample_to_file):
        gan = self.gan

        sess = gan.session
        config = gan.config
        if self.xs is None:
            self.xs = [sess.run([gan.inputs.x, gan.latent.sample]) for i in range(self.samples)]

        current_g = sess.run(gan.trainer.all_g_vars)
        
        stacks = []
        def _samples():
            n = 1
            cs = []
            for i in range(self.samples):
                ts = [gan.inputs.x, gan.latent.z]
                vs = self.xs[i]
                feed_dict = {}
                for t,v in zip(ts, vs):
                    feed_dict[t]=v
                cs.append(sess.run(gan.generator.sample,feed_dict))
                #cs.append(sess.run(gan.autoencoded_x,feed_dict))
            return np.vstack(cs)

        stacks.append(_samples())
        for sg in gan.trainer.sgs:
            gan.trainer.assign_g(sg)
            stacks.append(_samples())
        for i in range((gan.trainer.config.nash_memory_size or 10) - len(stacks)+1):
            stacks.append(_samples())

        gan.trainer.assign_g(current_g)

        images = np.vstack([np.hstack(s) for s in stacks])

        self.plot(images, path, sample_to_file)
        return [{'image': path, 'label': 'tiled x sample'}]



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from hypergan.train_hooks.base_train_hook import BaseTrainHook
from operator import itemgetter
from tensorflow.python.framework import ops
from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import state_ops
from tensorflow.python.training import optimizer
from torch.autograd import Variable
from torch.autograd import grad as torch_grad
import hyperchamber as hc
import inspect
import numpy as np
import torch
import torch.nn as nn

class LearningRateDropoutTrainHook(BaseTrainHook):
  """ https://arxiv.org/abs/1912.00144 """
  def __init__(self, gan=None, config=None, trainer=None):
      super().__init__(config=config, gan=gan, trainer=trainer)
      self.ones = self.gan.configurable_param(self.config.ones or 1.0)
      self.zeros = self.gan.configurable_param(self.config.zeros or 0.0)
      self.dropout = self.gan.configurable_param(self.config.dropout or 0.9)

  def forward(self):
      return [None, None]

  def gradients(self, d_grads, g_grads):
      d_ones = [torch.ones_like(_g) for _g in d_grads]
      g_ones = [torch.ones_like(_g) for _g in g_grads]
      d_zeros = [torch.zeros_like(_g) for _g in d_grads]
      g_zeros = [torch.zeros_like(_g) for _g in g_grads]
      da = [torch.where((torch.rand_like(_d_grads)- (1.0-self.dropout)) < 0, _o, _z) for _d_grads, _o, _z in zip(d_grads, d_ones, d_zeros)]
      ga = [torch.where((torch.rand_like(_g_grads)- (1.0-self.dropout)) < 0, _o, _z) for _g_grads, _o, _z in zip(g_grads, g_ones, g_zeros)]

      if self.config.skip_d is None:
          d_grads = [_a * _grad for _a, _grad in zip(da, d_grads)]
      if self.config.skip_g is None:
          g_grads = [_a * _grad for _a, _grad in zip(ga, g_grads)]
      #def count_params(variables):
      #  return np.sum([np.prod(self.ops.shape(t)) for t in variables ])
      #self.gan.add_metric('dropout-one', ones)
      #self.gan.add_metric('dropout-perc', dropout)
      #self.gan.add_metric('dropout', sum([self.ops.squash((1.0-_a/ones), tf.reduce_sum) for _a in da]) / count_params(da))

      return [d_grads, g_grads]



from .base_gan import BaseGAN
from hyperchamber import Config
from hypergan.discriminators import *
from hypergan.distributions import *
from hypergan.gan_component import ValidationException, GANComponent
from hypergan.generators import *
from hypergan.inputs import *
from hypergan.samplers import *
from hypergan.trainers import *
import copy
import hyperchamber as hc
import hypergan as hg
import importlib
import json
import numpy as np
import os
import sys
import time
import torch
import uuid

class StandardGAN(BaseGAN):
    """ 
    Standard GANs consist of:
    
    *required to sample*
    
    * latent
    * generator
    * sampler

    *required to train*

    * discriminator
    * loss
    * trainer
    """
    def __init__(self, *args, **kwargs):
        self.discriminator = None
        self.latent = None
        self.generator = None
        self.loss = None
        self.trainer = None
        self.features = []
        BaseGAN.__init__(self, *args, **kwargs)
        self.x = self.inputs.next()

    def build(self):
        torch.onnx.export(self.generator, torch.randn(*self.latent.z.shape, device='cuda'), "generator.onnx", verbose=True, input_names=["latent"], output_names=["generator"])

    def required(self):
        return "generator".split()

    def create(self):
        config = self.config

        self.latent = self.create_component("latent")
        self.generator = self.create_component("generator", input=self.latent)
        self.discriminator = self.create_component("discriminator")
        self.loss = self.create_component("loss")
        self.trainer = self.create_component("trainer")

    def forward_discriminator(self, inputs):
        return self.discriminator(inputs[0])

    def forward_pass(self):
        self.x = self.inputs.next()
        g = self.generator(self.latent.next())
        self.g = g
        d_real = self.forward_discriminator([self.x])
        d_fake = self.forward_discriminator([g])
        self.d_fake = d_fake
        self.d_real = d_real
        return d_real, d_fake

    def input_nodes(self):
        "used in hypergan build"
        return [
        ]

    def output_nodes(self):
        "used in hypergan build"
        return [
        ]

    def discriminator_components(self):
        return [self.discriminator]

    def generator_components(self):
        return [self.generator]

    def discriminator_fake_inputs(self, discriminator_index=0):
        return [self.g]

    def discriminator_real_inputs(self, discriminator_index=0):
        if hasattr(self, 'x'):
            return [self.x]
        else:
            return [self.inputs.next()]




import hyperchamber as hc
import numpy as np
import torch
from .base_distribution import BaseDistribution

from ..gan_component import ValidationException
from torch.distributions import uniform

TINY=1e-12

class TruncatedNormalDistribution(BaseDistribution):
    def __init__(self, gan, config):
        BaseDistribution.__init__(self, gan, config)
        self.current_channels = config["z"]
        self.current_width = 1
        self.current_height = 1
        self.current_input_size = config["z"]
        batch_size = gan.batch_size()
        self.z = torch.Tensor(batch_size, self.current_input_size).cuda()

    def create(self):
        pass

    def required(self):
        return "".split()

    def validate(self):
        errors = BaseDistribution.validate(self)
        #if(self.config.z is not None and int(self.config.z) % 2 != 0):
        #    errors.append("z must be a multiple of 2 (was %2d)" % self.config.z)
        return errors

    # https://discuss.pytorch.org/t/implementing-truncated-normal-initializer/4778/15
    def truncated_normal_(self, tensor, mean=0, std=1):
        size = tensor.shape
        tmp = tensor.new_empty(size + (4,)).normal_()
        valid = (tmp < 2) & (tmp > -2)
        ind = valid.max(-1, keepdim=True)[1]
        tensor.data.copy_(tmp.gather(-1, ind).squeeze(-1))
        tensor.data.mul_(std).add_(mean)

    def sample(self):
        self.truncated_normal_(self.z)
        return self.z

    def next(self):
        self.instance = self.sample()
        return self.instance



#From https://gist.github.com/EndingCredits/b5f35e84df10d46cfa716178d9c862a3
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import state_ops
from tensorflow.python.framework import ops
from tensorflow.python.training import optimizer
import hyperchamber as hc
import numpy as np
import inspect
from operator import itemgetter
from hypergan.train_hooks.base_train_hook import BaseTrainHook

EPS=1e-8

class ProgressCompressTrainHook(BaseTrainHook):
  """https://arxiv.org/pdf/1805.06370v2.pdf"""
  def __init__(self, gan=None, config=None, trainer=None, name="ProgressCompressTrainHook"):
    super().__init__(config=config, gan=gan, trainer=trainer, name=name)
    d_loss = []

    self.x = tf.Variable(tf.zeros_like(gan.inputs.x))
    self.g = tf.Variable(tf.zeros_like(gan.generator.sample))

    stacked = tf.concat([self.x, self.g], axis=0)
    self.kb = gan.create_component(config["knowledge_base"], name="knowledge_base", input=stacked)
    self.assign_x = tf.assign(self.x, gan.inputs.x)
    self.assign_g = tf.assign(self.g, gan.generator.sample)
    self.re_init_d = [d.initializer for d in gan.discriminator.variables()]
    gan.hack = self.g

    self.assign_knowledge_base = []
    for c in gan.components:
        if hasattr(c, 'knowledge_base'):
            for name, net in c.knowledge_base:
                assign = self.kb.named_layers[name]
                if self.ops.shape(assign)[0] > self.ops.shape(net)[0]:
                    assign = tf.slice(assign,[0 for i in self.ops.shape(net)] , [self.ops.shape(net)[0]]+self.ops.shape(assign)[1:])
                self.assign_knowledge_base.append(tf.assign(net, assign))

    def kl_divergence(_p, _q):
        return tf.reduce_sum(_p * tf.log(_p/(_q+EPS)+EPS))

    if self.config.method == 'gan':
        bs = gan.batch_size()
        real = tf.reshape(gan.loss.sample[:2], [bs, 2])
        fake = tf.reshape(self.kb.sample, [bs, 2])
        gl = tf.concat([real, fake], axis=0)
        self.kb_d = gan.create_component(self.config.knowledge_discriminator, name="kb_d", input=gl)
        self.kb_loss = self.gan.create_loss(self.kb_d)
        self.loss = self.kb_loss.sample[0] + self.kb_loss.sample[1]
        variables = self.kb_d.variables()
        variables += self.kb.variables()
    else:
        self.kb_loss = self.gan.create_component(gan.config.loss, discriminator=self.kb, split=2)

        self.loss = kl_divergence(gan.loss.sample[0], self.kb_loss.sample[0])
        if self.config.kl_on_d_fake:
            self.loss = kl_divergence(gan.loss.d_real, self.kb_loss.d_real)
        if self.config.kl_on_g:
            self.loss += kl_divergence(gan.loss.sample[1], self.kb_loss.sample[1])
        variables = self.kb.variables()
    config["optimizer"]["loss"] = self.loss
    self.optimizer = self.gan.create_optimizer(config["optimizer"])
    grads = tf.gradients(self.loss, variables)
    apply_vec = list(zip(grads, variables)).copy()
    self.optimize_t = self.optimizer.apply_gradients(apply_vec, global_step=gan.global_step)

    self.gan.add_metric('kbl', self.kb_loss.sample[0])
    self.gan.add_metric('kblg', self.kb_loss.sample[1])
    self.gan.add_metric('compress', self.loss)

  def losses(self):
      return [None, None]

  def after_step(self, step, feed_dict):
    if step % (self.config.step_count or 1) != 0:
      return
    # compress
    for i in range(self.config.night_steps or 1):
        self.gan.session.run(self.optimize_t)
    if self.config.reinitialize_every:
        if step % (self.config.reinitialize_every)==0 and step > 0:
            print("Reinitializing active D")
            self.gan.session.run(self.re_init_d)

  def before_step(self, step, feed_dict):
    self.gan.session.run([self.assign_x, self.assign_g]+ self.assign_knowledge_base)




from hypergan.viewer import GlobalViewer
import hyperchamber as hc
import numpy as np
import inspect
from hypergan.gan_component import ValidationException, GANComponent
from torch.autograd import grad as torch_grad
from operator import itemgetter
from hypergan.train_hooks.base_train_hook import BaseTrainHook

from hypergan.viewer import GlobalViewer

class MatchSupportTrainHook(BaseTrainHook):
    """ 
      Makes d_fake and d_real balance by running an optimizer.
    """
    def __init__(self, gan=None, config=None, trainer=None):
        super().__init__(config=config, gan=gan, trainer=trainer)

    def before_step(self, step, feed_dict, depth=0):
        defn = self.config.optimizer.copy()
        klass = GANComponent.lookup_function(None, defn['class'])
        del defn["class"]
        self.optimizer = klass(self.gan.generator.parameters(), **defn)

        for i in range(self.config.steps or 1):
            self.optimizer.zero_grad()
            fake = self.gan.discriminator(self.gan.generator(self.gan.latent.instance)).mean()
            real = self.gan.discriminator(self.gan.inputs.sample).mean()
            loss = self.gan.loss.forward_adversarial_norm(real, fake)
            if loss == 0.0:
                if self.config.verbose:
                    print("[match support] No loss")
                break
            move = torch_grad(outputs=loss, inputs=self.gan.g_parameters(), retain_graph=True, create_graph=True)

            if self.config.regularize:
                move = torch_grad(outputs=(loss+sum([m.abs().sum() for m in move])), inputs=self.gan.g_parameters(), retain_graph=True, create_graph=True)
            for p, g in zip(self.gan.g_parameters(), move):
                if p._grad is not None:
                    p._grad.copy_(g)
                else:
                    pass
                    #print("Missing g")
            self.optimizer.step()
            if self.config.verbose:
                print("[match support]", i, "loss", loss.item())
            if self.config.loss_threshold and loss < self.config.loss_threshold:
                if self.config.info:
                    print("[match support] loss_threshold steps", i, "loss", loss.item())
                break
        if self.config.info and i == self.config.steps-1:
            print("[match support] loss_threshold steps", i, "loss", loss.item())



#From https://gist.github.com/EndingCredits/b5f35e84df10d46cfa716178d9c862a3
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import state_ops
from tensorflow.python.framework import ops
from tensorflow.python.training import optimizer
import hyperchamber as hc
import inspect

class PotentialOptimizer(optimizer.Optimizer):
  def __init__(self, learning_rate=0.001, p=0.1, gan=None, config=None, use_locking=False, name="CurlOptimizer", optimizer=None, rho=1, beta=1, gamma=1):
    super().__init__(use_locking, name)
    self.gan = gan
    self.config = config
    self.name = name
    self.optimizer = self.gan.create_optimizer(optimizer)
 
  def _prepare(self):
    super()._prepare()
    self.optimizer._prepare()

  def _create_slots(self, var_list):
    super()._create_slots(var_list)
    self.optimizer._create_slots(var_list)

  def _apply_dense(self, grad, var):
    return self.optimizer._apply_dense(grad, var)

  def apply_gradients(self, grads_and_vars, global_step=None, name=None):
    var_list = [ v for _,v in grads_and_vars]
    grad_list = [ g for g,_ in grads_and_vars]
    d_vars = []
    g_vars = []
    for grad,var in grads_and_vars:
        if var in self.gan.d_vars():
            d_vars += [var]
        elif var in self.gan.g_vars():
            g_vars += [var]
        else:
            raise("Couldn't find var in g_vars or d_vars")

    with ops.init_scope():
        [self._get_or_make_slot(v, v, "potential", self._name) for v in var_list]
        [self._get_or_make_slot(v, v, "start", self._name) for v in var_list]
        self.optimizer._create_slots([v for g,v in grads_and_vars])
        for name in self.optimizer.get_slot_names():
            for var in self.optimizer.variables():
                self._zeros_slot(var, "potential", self.name)
                self._zeros_slot(var, "start", self.name)

    self._prepare()
    potential_slots = [self.get_slot(v, "potential") for v in var_list]
    start_slots = [self.get_slot(v, "start") for v in var_list]
    for name in self.optimizer.get_slot_names():
        for var in self.optimizer.variables():
            var_list += [var]
            potential_slots += [self._zeros_slot(var, "potential", self.name)]
            start_slots += [self._zeros_slot(var, "start", self.name)]

    consensus_reg = [tf.square(g) for g in grad_list[:len(d_vars)] if g is not None]
    Jgrads = tf.gradients(consensus_reg, d_vars) + [tf.zeros_like(g) for g in g_vars]

    shapes = [self.gan.ops.shape(l) for l in grad_list]
    u = [tf.reshape(l, [-1]) for l in grad_list[:len(d_vars)]]
    v = [tf.reshape(l, [-1]) if l is not None else tf.reshape(tf.zeros_like(v), [-1]) for l,v in zip(Jgrads[:len(d_vars)], d_vars)]
    
    def proj(u, v,shape):
        bounds = [-1.0,1.0]
        if self.config.ortho_bounds:
            bounds = self.config.ortho_bounds
        dot = tf.tensordot(v, u, 1) / (tf.square(u)+1e-8)
        dot = tf.maximum(bounds[0], dot)
        dot = tf.minimum(bounds[1], dot)
        dot = dot * u
        dot = tf.reshape(dot, shape)
        return dot
    projs = [proj(_u, _v, _s) for _u, _v, _s in zip(u, v, shapes)]
    if self.config.formulation == 'g-p,-p':
        h_grads = [grad-proj for grad, proj in zip(grad_list, projs)]
    elif self.config.formulation == 'g-p,g-p':
        h_grads = [grad-proj for grad, proj in zip(grad_list, projs)]
    elif self.config.formulation == 'g-p,g+p':
        h_grads = [grad-proj for grad, proj in zip(grad_list, projs)]
    elif self.config.formulation == 'p,-p':
        h_grads = [proj for grad, proj in zip(grad_list, projs)]
    elif self.config.formulation == 'g-p,p':
        h_grads = [grad-proj for grad, proj in zip(grad_list, projs)]
    else:
        h_grads = [grad+proj for grad, proj in zip(grad_list, projs)]
    step1_grads = h_grads + grad_list[len(d_vars):]

    step1 = list(zip(step1_grads, var_list))

    mlam = self.config.merge_lambda or 0.5

    op1 = tf.group(*[tf.assign(w, v) for w,v in zip(start_slots, var_list)]) # store variables
    with tf.get_default_graph().control_dependencies([op1]):
        op2 = self.optimizer.apply_gradients(step1.copy(), global_step=global_step, name=name)
        with tf.get_default_graph().control_dependencies([op2]):
            op3 = tf.group(*[tf.assign(w, v) for w,v in zip(potential_slots, var_list)]) # store variables
            with tf.get_default_graph().control_dependencies([op3]):
                op4 = tf.group(*[tf.assign(w, v) for w,v in zip(var_list, start_slots)])
                with tf.get_default_graph().control_dependencies([op4]):
                    if self.config.formulation == 'p,-p':
                        p_grads = [-proj for _g, proj in zip(grad_list, projs)]
                    elif self.config.formulation == 'g-p,g-p':
                        p_grads = [_g-proj for _g, proj in zip(grad_list, projs)]
                    elif self.config.formulation == 'g-p,g+p':
                        p_grads = [_g+proj for _g, proj in zip(grad_list, projs)]
                    elif self.config.formulation == 'g-p,-p':
                        p_grads = [-proj for _g, proj in zip(grad_list, projs)]
                    elif self.config.formulation == 'g-p,p':
                        p_grads = [proj for _g, proj in zip(grad_list, projs)]
                    else:
                        p_grads = [proj for _g, proj in zip(grad_list, projs)]
                    step2_grads = p_grads + grad_list[len(d_vars):]
                    step2 = list(zip(step2_grads, var_list))
                    op5 = self.optimizer.apply_gradients(step2.copy(), global_step=global_step, name=name)
                    with tf.get_default_graph().control_dependencies([op5]):
                        if self.config.ema:
                            op6 = tf.group(*[tf.assign(w, self.config.ema*start + (1.0-self.config.ema)*(mlam*h+(1-mlam)*p)) for start,w,h,p in zip(start_slots, var_list, var_list, potential_slots)])
                        else:
                            op6 = tf.group(*[tf.assign(w, mlam*h+(1-mlam)*p) for w,h,p in zip(var_list, var_list, potential_slots)])
                        with tf.get_default_graph().control_dependencies([op6]):
                            return tf.no_op()


  def _apply_sparse(self, grad, var):
    raise NotImplementedError("Sparse gradient updates are not supported.")
  def variables(self):
      return super().variables() + self.optimizer.variables()



#From https://gist.github.com/EndingCredits/b5f35e84df10d46cfa716178d9c862a3
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import torch
from torch.autograd import Variable
from torch.autograd import grad as torch_grad
import torch.nn as nn

import hyperchamber as hc
import numpy as np
import inspect
from operator import itemgetter
from hypergan.train_hooks.base_train_hook import BaseTrainHook

class GradientPenaltyTrainHook(BaseTrainHook):
  def __init__(self, gan=None, config=None, trainer=None):
      super().__init__(config=config, gan=gan, trainer=trainer)
      flex = 1
      if self.config.flex is not None:
          flex = self.config.flex
      self.flex = self.gan.configurable_param(flex)
      self.relu = nn.ReLU()

  def forward(self, d_loss, g_loss):
    x = self.gan.inputs.sample
    g = self.gan.generator_sample

    if self.config.input == 'x':
        inp = x
    elif self.config.input == 'g':
        inp = g
    else:
        alpha = torch.rand(self.gan.batch_size(), 1, 1, 1).cuda()
        inp = alpha * x + (1 - alpha) * g
    #inp = inp(interpolated, requires_grad=True).cuda()

    gamma= self.gan.configurable_param(self.config['lambda'] or 1)

    d = self.gan.discriminator(inp, context=self.gan.gp_context)
    parameters = list(self.gan.d_parameters())#+ [interpolated]

    grad = torch_grad(outputs=d.mean(), inputs=parameters, create_graph=True, retain_graph=True)

    grad = [_g.view(-1) for _g in grad]
    if self.config.type == "relu":
        grad_loss = [(self.relu(torch.sqrt(torch.sum(_g**2, dim=0) +1e-12) - self.flex) ** 2).mean() for _g in grad]
    else:
        grad_loss = [((torch.sqrt(torch.sum(_g**2, dim=0) +1e-12) - self.flex) ** 2).mean() for _g in grad]
    grad_loss = sum(grad_loss)

    loss = gamma*grad_loss
    return [loss, loss]

  def after_step(self, step, feed_dict):
    pass

  def before_step(self, step, feed_dict):
    pass



import math
import torch
from torch.optim import Optimizer

class GigaWolfOptimizer(Optimizer):
  def __init__(self, params, optimizer, optimizer2):
    defaults = dict()

    self.optimizer = optimizer
    self.optimizer2 = optimizer2
    super(GigaWolfOptimizer, self).__init__(params, defaults)

   def step(self, closure=None):
    """Performs a single optimization step across the two optimizers and uses GigaWolf.

    Arguments:
        closure (callable, optional): A closure that reevaluates the model
            and returns the loss.
    """



import numpy as np
from PIL import Image
from hypergan.viewer import GlobalViewer

class BaseSampler:
    def __init__(self, gan, samples_per_row=8, session=None):
        self.gan = gan
        self.samples_per_row = samples_per_row

    def _sample(self):
        raise "raw _sample method called.  You must override this"

    def compatible_with(gan):
        return False

    def sample(self, path, save_samples):
        gan = self.gan

        sample = self._sample()

        stacks = []
        for key, data in sample:
            data = data.cpu().permute(0,2,3,1).detach().numpy()
            #print(len(data))
            slots = min(gan.batch_size(), self.samples_per_row)
            slots = min(slots, np.shape(data)[0])
            #print('slots', slots, range(np.shape(data)[0]))
            #print([np.shape(data[i*slots:i*slots+slots]) for i in range(np.shape(data)[0]//slots)])
            #stacks += [np.vstack(data[i*slots:i*slots+slots]) for i in range(np.shape(data)[0]//slots)]
            stacks += [np.hstack(data[i*slots:i*slots+slots]) for i in range(np.shape(data)[0]//slots)]
        #print([np.shape(stack) for stack in stacks])
        #sample_data = np.hstack(stacks)
        sample_data = np.vstack(stacks)
        image = self.plot(sample_data, path, save_samples)
        sample_name = 'generator'
        samples = [[sample_data, sample_name]]

        return [{'image':path, 'label':'sample'} for sample_data, sample_filename in samples]

    def plot(self, image, filename, save_sample, regularize=True):
        """ Plot an image."""
        if regularize:
            image = np.minimum(image, 1)
            image = np.maximum(image, -1)
        image = np.squeeze(image)
        if np.shape(image)[2] == 4:
            fmt = "RGBA"
        else:
            fmt = "RGB"
        # Scale to 0..255.
        imin, imax = -1.0, 1.0
        image = (image - imin) * 255. / (imax - imin) + .5
        image = image.astype(np.uint8)
        if save_sample:
            try:
                Image.fromarray(image, fmt).save(filename)
            except Exception as e:
                print("Warning: could not sample to ", filename, ".  Please check permissions and make sure the path exists")
                print(e)

        GlobalViewer.update(self.gan, image)
        return image

    def plot_image(self, image, filename, save_sample, regularize=True):
        """ Plot an image from an external source."""
        if np.shape(image)[2] == 4:
            fmt = "RGBA"
        else:
            fmt = "RGB"
        if save_sample:
            try:
                Image.fromarray(image, fmt).save(filename)
            except Exception as e:
                print("Warning: could not sample to ", filename, ".  Please check permissions and make sure the path exists")
                print(e)

        GlobalViewer.update(self.gan, image)
        return image



"""AMSGrad for TensorFlow."""

from tensorflow.python.framework import ops
from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import resource_variable_ops
from tensorflow.python.ops import state_ops
from tensorflow.python.ops import variable_scope
from tensorflow.python.training import optimizer


class AMSGradOptimizer(optimizer.Optimizer):
    def __init__(self, learning_rate=0.01, beta1=0.9, beta2=0.99, epsilon=1e-8, use_locking=False, name="AMSGrad"):
        super(AMSGradOptimizer, self).__init__(use_locking, name)
        self._lr = learning_rate
        self._beta1 = beta1
        self._beta2 = beta2
        self._epsilon = epsilon

        self._lr_t = None
        self._beta1_t = None
        self._beta2_t = None
        self._epsilon_t = None

        self._beta1_power = None
        self._beta2_power = None
        self._variables = []

    def variables(self):
        return self._variables

    def _create_slots(self, var_list):
        first_var = min(var_list, key=lambda x: x.name)

        create_new = self._beta1_power is None

        if create_new:
            with ops.colocate_with(first_var):
                self._beta1_power = variable_scope.variable(self._beta1, name="beta1_power", trainable=False)
                self._beta2_power = variable_scope.variable(self._beta2, name="beta2_power", trainable=False)
                self._variables += [self._beta1_power]
                self._variables += [self._beta2_power]
        # Create slots for the first and second moments.
        for v in var_list :
            self._variables += [self._zeros_slot(v, "m", self._name)]
            self._variables += [self._zeros_slot(v, "v", self._name)]
            self._variables += [self._zeros_slot(v, "vhat", self._name)]

    def _prepare(self):
        self._lr_t = ops.convert_to_tensor(self._lr)
        self._beta1_t = ops.convert_to_tensor(self._beta1)
        self._beta2_t = ops.convert_to_tensor(self._beta2)
        self._epsilon_t = ops.convert_to_tensor(self._epsilon)

    def _apply_dense(self, grad, var):
        beta1_power = math_ops.cast(self._beta1_power, var.dtype.base_dtype)
        beta2_power = math_ops.cast(self._beta2_power, var.dtype.base_dtype)
        lr_t = math_ops.cast(self._lr_t, var.dtype.base_dtype)
        beta1_t = math_ops.cast(self._beta1_t, var.dtype.base_dtype)
        beta2_t = math_ops.cast(self._beta2_t, var.dtype.base_dtype)
        epsilon_t = math_ops.cast(self._epsilon_t, var.dtype.base_dtype)

        lr = (lr_t * math_ops.sqrt(1 - beta2_power) / (1 - beta1_power))

        # m_t = beta1 * m + (1 - beta1) * g_t
        m = self.get_slot(var, "m")
        m_scaled_g_values = grad * (1 - beta1_t)
        m_t = state_ops.assign(m, beta1_t * m + m_scaled_g_values, use_locking=self._use_locking)

        # v_t = beta2 * v + (1 - beta2) * (g_t * g_t)
        v = self.get_slot(var, "v")
        v_scaled_g_values = (grad * grad) * (1 - beta2_t)
        v_t = state_ops.assign(v, beta2_t * v + v_scaled_g_values, use_locking=self._use_locking)

        # amsgrad
        vhat = self.get_slot(var, "vhat")
        vhat_t = state_ops.assign(vhat, math_ops.maximum(v_t, vhat))
        v_sqrt = math_ops.sqrt(vhat_t)

        var_update = state_ops.assign_sub(var, lr * m_t / (v_sqrt + epsilon_t), use_locking=self._use_locking)
        return control_flow_ops.group(*[var_update, m_t, v_t, vhat_t])

    def _resource_apply_dense(self, grad, var):
        var = var.handle
        beta1_power = math_ops.cast(self._beta1_power, grad.dtype.base_dtype)
        beta2_power = math_ops.cast(self._beta2_power, grad.dtype.base_dtype)
        lr_t = math_ops.cast(self._lr_t, grad.dtype.base_dtype)
        beta1_t = math_ops.cast(self._beta1_t, grad.dtype.base_dtype)
        beta2_t = math_ops.cast(self._beta2_t, grad.dtype.base_dtype)
        epsilon_t = math_ops.cast(self._epsilon_t, grad.dtype.base_dtype)

        lr = (lr_t * math_ops.sqrt(1 - beta2_power) / (1 - beta1_power))

        # m_t = beta1 * m + (1 - beta1) * g_t
        m = self.get_slot(var, "m").handle
        m_scaled_g_values = grad * (1 - beta1_t)
        m_t = state_ops.assign(m, beta1_t * m + m_scaled_g_values, use_locking=self._use_locking)

        # v_t = beta2 * v + (1 - beta2) * (g_t * g_t)
        v = self.get_slot(var, "v").handle
        v_scaled_g_values = (grad * grad) * (1 - beta2_t)
        v_t = state_ops.assign(v, beta2_t * v + v_scaled_g_values, use_locking=self._use_locking)

        # amsgrad
        vhat = self.get_slot(var, "vhat").handle
        vhat_t = state_ops.assign(vhat, math_ops.maximum(v_t, vhat))
        v_sqrt = math_ops.sqrt(vhat_t)

        var_update = state_ops.assign_sub(var, lr * m_t / (v_sqrt + epsilon_t), use_locking=self._use_locking)
        return control_flow_ops.group(*[var_update, m_t, v_t, vhat_t])

    def _apply_sparse_shared(self, grad, var, indices, scatter_add):
        beta1_power = math_ops.cast(self._beta1_power, var.dtype.base_dtype)
        beta2_power = math_ops.cast(self._beta2_power, var.dtype.base_dtype)
        lr_t = math_ops.cast(self._lr_t, var.dtype.base_dtype)
        beta1_t = math_ops.cast(self._beta1_t, var.dtype.base_dtype)
        beta2_t = math_ops.cast(self._beta2_t, var.dtype.base_dtype)
        epsilon_t = math_ops.cast(self._epsilon_t, var.dtype.base_dtype)

        lr = (lr_t * math_ops.sqrt(1 - beta2_power) / (1 - beta1_power))

        # m_t = beta1 * m + (1 - beta1) * g_t
        m = self.get_slot(var, "m")
        m_scaled_g_values = grad * (1 - beta1_t)
        m_t = state_ops.assign(m, m * beta1_t, use_locking=self._use_locking)
        with ops.control_dependencies([m_t]):
            m_t = scatter_add(m, indices, m_scaled_g_values)

        # v_t = beta2 * v + (1 - beta2) * (g_t * g_t)
        v = self.get_slot(var, "v")
        v_scaled_g_values = (grad * grad) * (1 - beta2_t)
        v_t = state_ops.assign(v, v * beta2_t, use_locking=self._use_locking)
        with ops.control_dependencies([v_t]):
            v_t = scatter_add(v, indices, v_scaled_g_values)

        # amsgrad
        vhat = self.get_slot(var, "vhat")
        vhat_t = state_ops.assign(vhat, math_ops.maximum(v_t, vhat))
        v_sqrt = math_ops.sqrt(vhat_t)
        var_update = state_ops.assign_sub(var, lr * m_t / (v_sqrt + epsilon_t), use_locking=self._use_locking)
        return control_flow_ops.group(*[var_update, m_t, v_t, vhat_t])

    def _apply_sparse(self, grad, var):
        return self._apply_sparse_shared(
            grad.values, var, grad.indices,
            lambda x, i, v: state_ops.scatter_add(  # pylint: disable=g-long-lambda
                x, i, v, use_locking=self._use_locking))

    def _resource_scatter_add(self, x, i, v):
        with ops.control_dependencies(
                [resource_variable_ops.resource_scatter_add(x.handle, i, v)]):
            return x.value()

    def _resource_apply_sparse(self, grad, var, indices):
        return self._apply_sparse_shared(
            grad, var, indices, self._resource_scatter_add)

    def _finish(self, update_ops, name_scope):
        # Update the power accumulators.
        with ops.control_dependencies(update_ops):
            with ops.colocate_with(self._beta1_power):
                update_beta1 = self._beta1_power.assign(
                    self._beta1_power * self._beta1_t,
                    use_locking=self._use_locking)
                update_beta2 = self._beta2_power.assign(
                    self._beta2_power * self._beta2_t,
                    use_locking=self._use_locking)
        return control_flow_ops.group(*update_ops + [update_beta1, update_beta2],
                                      name=name_scope)
